from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    import mince.types
    import polars as pl


def get_polars_type_map() -> dict[mince.types.ColumnType, str]:
    return {
        pl.Int8: 'Int8',
        pl.Int16: 'Int16',
        pl.Int32: 'Int32',
        pl.Int64: 'Int64',
        pl.UInt8: 'UInt8',
        pl.UInt16: 'UInt16',
        pl.UInt32: 'UInt32',
        pl.UInt64: 'UInt64',
        pl.Float32: 'Float32',
        pl.Float64: 'Float64',
        pl.Decimal: 'Decimal',
        pl.Boolean: 'Boolean',
        pl.Binary: 'Binary',
        pl.String: 'String',
        pl.Datetime('ms'): "Datetime('ms')",
        pl.Duration('ms'): "Duration('ms')",
        pl.Date: 'Date',
        pl.Time: 'Time',
    }


def serialize_polars_schema(
    schema: pl.schema.Schema | dict[str, mince.types.ColumnType],
) -> list[tuple[str, str]]:
    """convert schema into json-compatible form"""
    import polars as pl

    if isinstance(schema, pl.schema.Schema):
        dict_schema: dict[str, mince.types.ColumnType] = dict(schema.items())
    elif isinstance(schema, dict):
        dict_schema = schema  # type: ignore
    else:
        raise Exception()

    type_map = get_polars_type_map()
    output = []
    for k, v in dict_schema.items():
        v_str = type_map.get(v)
        if v_str is not None:
            output.append((k, v_str))
        else:
            raise Exception('unknown column type: ' + str(v))

    return output


def deserialize_polars_schema(
    schema: typing.Sequence[typing.Sequence[str]],
) -> dict[str, mince.types.ColumnType]:
    """convert json-compatible schema into proper schema"""
    type_map = get_polars_type_map()
    reverse_type_map = {v: k for k, v in type_map.items()}
    output = {}
    for entry in schema:
        if len(entry) != 2:
            raise Exception('invalid schema structure')
        column_name, str_column_type = entry

        column_type = reverse_type_map.get(str_column_type)
        if column_type is None:
            raise Exception()
        output[column_name] = column_type

    return output



from __future__ import annotations

import argparse
import typing

import mince.ops
from .. import cli_helpers
from .. import cli_parsing

if typing.TYPE_CHECKING:
    from mince import types


def add_kill_command(
    subparsers: cli_helpers.Subparsers,
    include_dashboard_arg: bool,
    DashboardClass: typing.Type[mince.Dashboard] | None,
) -> None:
    parser = subparsers.add_parser(
        'kill',
        help='kill dashboard',
        formatter_class=cli_helpers.HelpFormatter,
    )
    parser.set_defaults(f_command=kill_command)
    if include_dashboard_arg:
        parser.add_argument('dashboard', help='dashboard name', nargs='?')
    parser.add_argument('--port', help='port number to kill dashboard of')
    parser.add_argument('--pid', help='pid to kill')
    cli_parsing.add_common_args(parser)


def kill_command(
    args: argparse.Namespace,
    dashboard: str | typing.Type[types.Dashboard] | None = None,
) -> None:
    if dashboard is None:
        dashboard = args.dashboard
    if isinstance(dashboard, str):
        mince.ops.kill_server(name=dashboard, port=args.port, pid=args.pid)
    else:
        raise NotImplementedError('cannot kill pure class')



from __future__ import annotations

from typing import Any, Literal, Protocol, Union
from typing_extensions import TypedDict, NotRequired
from .dashboards import Dashboard  # noqa: F401

import polars as pl
import tooltime


#
# # schema specification
#

Interval = Literal['point', 'day', 'week', 'month', 'year']

ColumnType = Union[pl.datatypes.classes.DataTypeClass, pl.datatypes.DataType]


class DataSchema(TypedDict):
    columns: dict[str, ColumnType]
    group_columns: list[str]


class DataSchemaPartial(TypedDict, total=False):
    columns: dict[str, ColumnType | None] | list[str]
    group_columns: list[str]


class MetricSpec(TypedDict):
    unit: str | None
    type: Literal[
        'point_in_time',
        'sum',
        'min',
        'max',
        'unique',
    ]


#
# # collection specification
#


class CollectKwargs(TypedDict):
    data_dir: str
    start_time: float
    end_time: float
    intervals: list[Interval]
    skip_incomplete_intervals: bool
    extra_kwargs: dict[str, Any]
    dry: bool
    verbose: int


class CollectKwargsPartial(TypedDict, total=False):
    data_dir: str | None
    start_time: tooltime.Timestamp | None
    end_time: tooltime.Timestamp | None
    intervals: list[Interval] | None
    skip_incomplete_intervals: bool | None
    extra_kwargs: dict[str, Any] | None
    extra_kwargs_update: dict[str, Any] | None
    dry: bool | None
    verbose: int | None


class CollectionJobSummary(TypedDict):
    mince_version: str
    dashboard_version: str
    dashboard_name: str
    dashboard_class: str
    job_start_time: float
    job_end_time: float
    data_names: list[str]
    collect_kwargs: CollectKwargs


class AsyncDfCollector(Protocol):
    async def __call__(
        self,
        start_time: int,
        end_time: int,
        interval: Interval,
        tags: dict[str, str],
    ) -> pl.DataFrame: ...


# # ui spec
#

StateDatum = Union[str, int, float]


class UiSpec(TypedDict):
    name: str
    title: str | None  # shown in browser tab title
    description: str | None  # shown in metadata and cli
    favicon_emoji: str | None  # single character str
    version: str
    metrics: dict[str, MetricSpec]
    groupings: list[str]
    default_state: dict[str, StateDatum | None]
    invalid_states: list[dict[str, StateDatum]]
    inputs: dict[str, InputSpec]
    submetrics: dict[str, list[str]]
    shortcuts: dict[str, ShortcutSpec]
    colors: dict[str, str]
    schema: DataSchema
    title_prefix: str | None
    title_infix: str | None
    title_postfix: str | None


class InputSpec(TypedDict):
    type: Literal['button', 'date']
    description: str
    default: str | None
    visibility: NotRequired[InputVisibility]
    button_options: NotRequired[list[str]]
    aliases: dict[str, StateDatum]


class InputVisibility(TypedDict):
    start_hidden: NotRequired[bool]
    hide_if: NotRequired[list[dict[str, Any]]]
    show_if: NotRequired[list[dict[str, Any]]]
    hide: NotRequired[bool]
    f: NotRequired[Any]  # function


class ShortcutSpec(TypedDict):
    action: Literal[
        'select',
        'cycle_next',
        'cycle_previous',
        'toggle_ui',
        'increment',
        'decrement',
    ]
    field: str | None
    value: NotRequired[str]
    help: NotRequired[str]


#
# # registry types
#

RegistryReference = Union['Registry', 'RegistryFile', str, None]


class Registry(TypedDict):
    mince_version: str
    dashboards: dict[str, 'RegistryEntry']


class RegistryFile(TypedDict):
    path: str | None
    validate: NotRequired[bool]
    create_if_dne: NotRequired[bool]


class RegistryEntry(TypedDict):
    name: str
    dashboard_class: str
    description: str | None
    collect_kwargs: CollectKwargsPartial


#
# # validation
#


def validate_typeddict(typed_dict: Any, td_class: type) -> None:
    annotations = td_class.__annotations__

    for field_name, field_type in annotations.items():
        if field_name not in typed_dict:
            raise ValueError(f'Missing required field: {field_name}')

        value = typed_dict[field_name]

        if field_type == pl.DataType:
            if not isinstance(value, pl.DataType):
                raise TypeError(f'Field {field_name} must be a polars DataType')
        elif isinstance(field_type, type):
            if not isinstance(value, field_type):
                raise TypeError(
                    f'Field {field_name} must be of type {field_type}'
                )
        else:
            # Handle more complex types (e.g., Union, List, etc.) here
            # This might require more sophisticated type checking
            pass

    # Check for extra fields
    extra_fields = set(typed_dict.keys()) - set(annotations.keys())
    if extra_fields:
        raise ValueError(f"Unexpected extra fields: {', '.join(extra_fields)}")



from __future__ import annotations

import typing

from mince import specs
from ... import dashboards
from .. import raw_figures

if typing.TYPE_CHECKING:
    import polars as pl
    import plotly.graph_objects as go  # type: ignore
    from mince import types


ui_interval_to_df_interval = {
    'date': 'day',
    'week': 'week',
    'month': 'month',
}


def create_treemap_fig(
    df: pl.DataFrame,
    state: dict[str, typing.Any],
    ui_spec: types.UiSpec,
    data_date_range: tuple[int, int],
    title: str | None = None,
) -> go.Figure:
    df = _prepare_treemap_data(df, state, ui_spec=ui_spec)
    styles = _get_treemap_styles(
        df, state, ui_spec, data_date_range, title=title
    )
    return raw_figures.create_treemap(df=df, **styles)


def _prepare_treemap_data(
    df: pl.DataFrame, state: dict[str, typing.Any], ui_spec: types.UiSpec
) -> pl.DataFrame:
    import polars as pl
    import tooltime

    # filter by metric
    metric = dashboards.get_df_metric(state, ui_spec)
    df = df.filter(pl.col.metric == metric)

    metric_type = ui_spec['metrics'][metric]['type']

    if metric_type == 'point_in_time':
        now = tooltime.timestamp_to_seconds(state['now']) * 1e3
        df = df.filter(pl.col.timestamp == now)

    elif metric_type == 'unique':
        # filter sample_interval
        ui_interval = state['sample_interval']
        interval = ui_interval_to_df_interval[ui_interval]
        df = df.filter(pl.col.interval == interval)

        # select the closest point to now
        time_values = df['timestamp'].sort().unique()
        now = tooltime.timestamp_to_seconds(state['now']) * 1e3
        time_index = time_values.cast(pl.Int64).search_sorted(now)
        if time_index == len(time_values):
            time_index = -1
        time = time_values[time_index]
        df = df.filter(pl.col.timestamp == time)

    elif metric_type in ['sum', 'min', 'max']:
        df = df.filter(pl.col.interval == 'day')

        # filter time window
        end_time = tooltime.timestamp_to_seconds(state['now'])
        if state['time_window'] == 'all':
            df = df.filter(pl.col.timestamp <= end_time * 1e3)
        else:
            window = tooltime.timelength_to_seconds(state['time_window'])
            start_time = end_time - window
            df = df.filter(
                pl.col.timestamp > start_time * 1e3,
                pl.col.timestamp <= end_time * 1e3,
            )

        if metric_type == 'sum':
            agg_function = pl.sum('value')
        elif metric_type == 'min':
            agg_function = pl.min('value')
        elif metric_type == 'max':
            agg_function = pl.max('value')
        else:
            raise Exception()

        # aggregate
        df = (
            df.filter(pl.col.metric == metric)
            .group_by(*ui_spec['schema']['group_columns'])
            .agg(
                agg_function,
                min_timestamp=pl.min('timestamp'),
                max_timestamp=pl.max('timestamp'),
            )
        )
    else:
        raise Exception('invalid metric_type for treemap')

    return df


def _get_treemap_styles(
    df: pl.DataFrame,
    state: dict[str, typing.Any],
    ui_spec: types.UiSpec,
    data_date_range: tuple[int, int],
    title: str | None = None,
) -> dict[str, typing.Any]:
    # create title
    if title is None:
        title = _create_tree_title(
            df=df,
            state=state,
            data_date_range=data_date_range,
            ui_spec=ui_spec,
        )

    # get prefix
    metric = dashboards.get_df_metric(state, ui_spec=ui_spec)
    metric_spec = ui_spec['metrics'][metric]
    if metric_spec['unit'] == '$':
        prefix = '$'
    else:
        prefix = None

    if len(ui_spec['groupings']) == 1:
        subgrouping = None
    elif len(ui_spec['groupings']) == 2:
        if state['grouping'] == ui_spec['groupings'][0]:
            subgrouping = ui_spec['groupings'][1]
        else:
            subgrouping = ui_spec['groupings'][0]
    elif len(ui_spec['groupings']) > 2:
        raise NotImplementedError('multiple subgroupings')
    else:
        raise Exception('no groupings specified')

    return {
        'title': title,
        'grouping': state['grouping'],
        'subgrouping': subgrouping,
        'metric': metric,
        'prefix': prefix,
        'colors': ui_spec.get('colors'),
    }


def _create_tree_title(
    df: pl.DataFrame,
    state: dict[str, typing.Any],
    data_date_range: tuple[int, int],
    ui_spec: types.UiSpec,
) -> str:
    import tooltime

    metric = dashboards.get_df_metric(state, ui_spec=ui_spec)
    metric_spec = ui_spec['metrics'][metric]

    title_prefix = ui_spec['title_prefix']
    title_infix = ui_spec['title_infix']
    title_postfix = ui_spec['title_postfix']

    if state['metric'] in ui_spec['submetrics']:
        submetric = specs.state_to_alias(
            state['submetric'], 'submetric', ui_spec
        ).lower()
        metric_str: str = submetric
    else:
        metric_str = specs.state_to_alias(
            state['metric'], 'metric', ui_spec
        ).lower()
    metric_str = metric_str.title()

    if title_infix is not None:
        metric_str = title_infix + ' ' + metric_str
    if title_postfix is not None:
        metric_str = metric_str + ' ' + title_postfix

    if metric_spec['type'] in ['sum', 'min', 'max']:
        title = metric_str
        min_date = tooltime.timestamp_to_date(df['min_timestamp'].min())  # type: ignore
        max_date = tooltime.timestamp_to_date(df['max_timestamp'].max())  # type: ignore
        title += ' over ' + '[' + min_date + ', ' + max_date + ']'
    elif metric_spec['type'] == 'unique':
        title = metric_str + ' during '
        dt = tooltime.timestamp_to_date(df['timestamp'].max())  # type: ignore
        if state['sample_interval'] == 'date':
            dt_str = 'day of ' + dt
        elif state['sample_interval'] == 'week':
            dt_str = 'week of ' + dt
        elif state['sample_interval'] == 'month':
            dt_str = 'month of ' + dt[:7]
        else:
            raise Exception(
                'invalid sampling_interval ' + str(state['sample_interval'])
            )
        title += dt_str
    else:
        title = metric_str
        title += ' ' + tooltime.timestamp_to_date(df['timestamp'].max())  # type: ignore

    if title_prefix is not None:
        title = title_prefix + title

    return title



from __future__ import annotations

import argparse
import typing

import mince.ops
from ... import storage
from .. import cli_helpers
from .. import cli_parsing

if typing.TYPE_CHECKING:
    from mince import types


def add_ls_command(
    subparsers: cli_helpers.Subparsers,
    include_dashboard_arg: bool,
    DashboardClass: typing.Type[mince.Dashboard] | None,
) -> None:
    parser = subparsers.add_parser(
        'ls',
        help='list registered or running dashboards',
        formatter_class=cli_helpers.HelpFormatter,
    )
    parser.set_defaults(f_command=ls_command)
    parser.add_argument(
        '-v',
        '--verbose',
        help='print extra info',
        action='store_true',
    )
    parser.add_argument(
        '--raw',
        help='print all info as raw json',
        action='store_true',
    )
    cli_parsing.add_common_args(parser)


def ls_command(args: argparse.Namespace, DashboardClass: typing.Any) -> None:
    print_registered_dashboards(args)
    print()
    print()
    print_running_dashboards(args)


def print_registered_dashboards(
    args: argparse.Namespace,
    dashboard: str | typing.Type[types.Dashboard] | None = None,
) -> None:
    import toolstr
    import rich

    rich.print('[bold][white]Registered dashboards:[/white][/bold]')
    registry = storage.load_registry_file(
        path=args.registry,
        validate=not args.no_validate,
    )
    if len(registry['dashboards']) == 0:
        print('[none]')
    else:
        if args.raw:
            import json

            print(json.dumps(list(registry.values()), indent=4, sort_keys=True))
            return
        rows = []
        for entry in registry['dashboards'].values():
            description = entry['description']
            row = [entry['name'], description]
            if args.verbose:
                row.append(entry['dashboard_class'])
            rows.append(row)
        labels = ['name', 'description']
        if args.verbose:
            labels.append('class')
        print()
        toolstr.print_table(
            rows,
            labels=labels,
            border='rgb(100,100,100)',
            column_styles={'name': 'green'},
            label_style='bold white',
            outer_gap=0,
            label_justify='left',
            column_justify='left',
        )


def print_running_dashboards(args: argparse.Namespace) -> None:
    import toolstr
    import tooltime
    import rich

    servers = mince.ops.get_running_servers()
    rich.print('[bold][white]Running dashboards:[/white][/bold]')
    if len(servers) == 0:
        print('[no running dashboards]')
    else:
        if args.raw:
            import json

            print(json.dumps(list(servers), indent=4, sort_keys=True))
            return
        print()
        rows = []
        for server in servers:
            row = [
                server['name'],
                'localhost:' + server['port'],
                'd = '
                + server['metadata']['dashboard_version']
                + '\n'
                + 'm = '
                + server['metadata']['mince_version'],
                server['metadata']['time_started'].replace(' ', '\n'),
                tooltime.timelength_to_clock_phrase(
                    float(tooltime.now())  # type: ignore
                    - tooltime.timestamp_to_seconds(
                        server['metadata']['time_started']
                    )
                ),
            ]
            if args.verbose:
                row.append(server['pid'])
            rows.append(row)

        labels = [
            'name',
            'url',
            'versions',
            'time\nstarted',
            'uptime',
        ]
        if args.verbose:
            labels.append('pid')
        toolstr.print_multiline_table(
            rows,
            labels=labels,
            border='rgb(100,100,100)',
            column_styles={'name': 'green'},
            label_style='bold white',
            column_justify='left',
            label_justify='left',
            outer_gap=0,
        )



from __future__ import annotations

from typing import Any, Sequence, TYPE_CHECKING


if TYPE_CHECKING:
    from dash import html, dcc, Dash  # type: ignore
    import tooltime
    from mince import types


def create_inputs(
    ui_spec: types.UiSpec,
    date_range: tuple[tooltime.Timestamp, tooltime.Timestamp],
) -> dict[str, html.Div]:
    from dash import html

    inputs: dict[str, html.Div] = {}
    for name, input_spec in ui_spec['inputs'].items():
        if input_spec['type'] == 'button':
            inputs[name] = create_button(name=name, input_spec=input_spec)
        elif input_spec['type'] == 'date':
            inputs[name] = create_date_picker(input_spec, date_range)
        else:
            raise Exception('invalid input type')
    return inputs


def create_button(name: str, input_spec: types.InputSpec) -> html.Div:
    start_hidden = input_spec.get('visibility', {}).get('start_hidden', False)
    return create_radio_group(
        name=name,
        options=input_spec['button_options'],
        default=input_spec['default'],
        description=input_spec['description'],
        start_hidden=start_hidden,
    )


def create_radio_group(
    name: str,
    options: Sequence[str],
    default: str | None = None,
    start_hidden: bool = False,
    description: str | None = None,
) -> html.Div:
    """create a radio button group"""
    from dash import html
    import dash_bootstrap_components as dbc  # type: ignore

    if default is None:
        default = options[0]

    if start_hidden:
        default_className = 'btn-group hidden'
    else:
        default_className = 'btn-group'

    return html.Div(
        [
            dbc.RadioItems(
                id=name + '-radio',
                className=default_className,
                inputClassName='btn-check',
                labelClassName='btn btn-outline-dark',
                labelCheckedClassName='active',
                options=[
                    {'label': option, 'value': option} for option in options
                ],
                value=default,
            ),
            dbc.Popover(
                description,
                target=name + '-radio',
                body=True,
                trigger='hover',
                placement='bottom',
            ),
        ],
        className='radio-group',
    )


def create_date_picker(
    input_spec: types.InputSpec,
    date_range: tuple[tooltime.Timestamp, tooltime.Timestamp],
) -> dcc.DatePickerSingle:
    from dash import html, dcc
    import dash_bootstrap_components as dbc
    import tooltime

    start_time, end_time = date_range
    min_date = tooltime.timestamp_to_datetime(start_time).strftime('%Y-%m-%d')
    max_date = tooltime.timestamp_to_datetime(end_time).strftime('%Y-%m-%d')
    return html.Div(
        [
            dcc.DatePickerSingle(
                id='date-picker',
                min_date_allowed=min_date,
                max_date_allowed=max_date,
                initial_visible_month=max_date,
                date=max_date,
                display_format='Y-MM-DD',
            ),
            dbc.Popover(
                input_spec['description'],
                target='date-picker',
                body=True,
                trigger='hover',
                placement='bottom',
            ),
        ],
        className='dash-bootstrap',
    )


def _prevent_button_focus(app: Dash) -> None:
    # this function is run to prevent radio buttons from being focus-able
    # otherwise, they conflict with the shortcuts used for the arrow keys
    from dash import Input, Output

    app.clientside_callback(
        """
        function(trigger) {
            if (!window.preventFocusApplied) {
                function preventFocus() {
                    document.querySelectorAll('.radio-group input[type="radio"]').forEach(radio => {
                        radio.addEventListener('focus', (e) => e.target.blur());
                    });
                }
                preventFocus();
                window.preventFocusApplied = true;
            }
            return '';
        }
        """,
        Output('prevent-focus-trigger', 'children'),
        Input('prevent-focus-trigger', 'children'),
    )


#
# # visibility
#


def compute_input_classes(
    state: dict[str, Any],
    ui_spec: types.UiSpec,
) -> dict[str, str]:
    classes = {True: 'btn-group', False: 'btn-group hidden'}
    return {
        name: classes[is_visible(state, input_spec)]
        for name, input_spec in ui_spec['inputs'].items()
    }


def is_visible(state: dict[str, Any], input_spec: types.InputSpec) -> bool:
    visibility = input_spec.get('visibility')
    if visibility is None:
        return True
    if visibility.get('hide'):
        return False
    if 'show_if' in visibility:
        if not any(is_match(state, match) for match in visibility['show_if']):
            return False
    elif 'hide_if' in input_spec['visibility']:
        if any(is_match(state, match) for match in visibility['hide_if']):
            return False
    return True


def is_match(display: dict[str, str], match: dict[str, str]) -> bool:
    for key, value in match.items():
        if display.get(key) != value:
            return False
    return True



from __future__ import annotations

import typing

from mince import specs
from . import dashboard_class
from . import dashboard_inputs
from . import dashboard_state
from . import dashboard_urls

if typing.TYPE_CHECKING:
    import flask
    from dash import Dash, Input, Output, State  # type: ignore
    import plotly.graph_objects as go  # type: ignore
    from mince import types


def _create_app(
    dashboard: dashboard_class.Dashboard,
    assets_folder: str | None = None,
    external_stylesheets: list[str] | None = None,
    pdb: bool = False,
) -> Dash:
    """create the dash.Dash object for a DashBoard"""
    from dash import Dash

    if external_stylesheets is None:
        import dash_bootstrap_components as dbc  # type: ignore

        external_stylesheets = [dbc.themes.BOOTSTRAP]

    # create app
    if assets_folder is None:
        assets_folder = get_default_assets_folder()
    app = Dash(
        dashboard.spec['name'],
        assets_folder=assets_folder,
        external_stylesheets=external_stylesheets,
    )
    app.title = dashboard.spec['name']
    favicon_emoji = dashboard.spec.get('favicon_emoji')
    if favicon_emoji is not None:
        app.index_string = app.index_string.replace(
            '{%favicon%}',
            '<link rel="icon" href="data:image/svg+xml,<svg xmlns=%22http://www.w3.org/2000/svg%22 viewBox=%220 0 100 100%22><text y=%22.9em%22 font-size=%2290%22>'  # noqa: E501
            + favicon_emoji
            + '</text></svg>">',
        )
    n_keydowns = 0

    # create inputs
    inputs = dashboard_inputs.create_inputs(
        dashboard.spec, dashboard.date_range
    )

    # create layout
    app.layout = dashboard.get_layout(inputs)

    # add serverside callbacks
    update_ui_args, update_chart_args = _compute_decorator_args(dashboard.spec)

    @app.callback(*update_ui_args)  # type: ignore
    def update_ui(*args) -> list[str | bool]:
        # parse input args
        (
            *raw_display_state,
            url,
            initial_load,
            _n_keydowns,
            shortcut,
            help_open,
            submetric_options,
            radio_group_visibility,
        ) = args
        display_state = dict(zip(list(inputs.keys()), raw_display_state))
        ui_state = {
            'initial_load': initial_load,
            'help_open': help_open,
            'submetric_options': submetric_options,
            'radio_group_visibility': radio_group_visibility,
        }

        # load from url
        if ui_state['initial_load']:
            dashboard_urls.parse_url(
                url, display_state, dashboard.spec['default_state']
            )
            ui_state['initial_load'] = False

        # process keyboard shortcuts
        nonlocal n_keydowns
        if _n_keydowns <= n_keydowns or shortcut is None:
            shortcut = None
        else:
            # help_open =
            dashboard.process_shortcuts(display_state, ui_state, shortcut)
        n_keydowns = _n_keydowns

        # specify submetrics list
        if display_state['metric'] in dashboard.spec['submetrics']:
            raw_options = dashboard.spec['submetrics'][display_state['metric']]
            button_options = []
            submetric_options = []
            for raw_option in raw_options:
                option = specs.state_to_alias(
                    raw_option, 'submetric', dashboard.spec
                )
                button_options.append(option)
                submetric_options.append({'label': option, 'value': option})
            ui_state['submetric_options'] = submetric_options
            dashboard.spec['inputs']['submetric']['button_options'] = (
                button_options
            )
            if display_state['submetric'] not in button_options:
                display_state['submetric'] = button_options[0]

        # based on most recent selection, fix any invalid button values
        dashboard_state._fix_invalid_inputs(
            display_state, shortcut, dashboard.spec
        )

        # update url based on radio button values
        url = dashboard_urls.create_new_url(
            display_state, defaults=dashboard.spec['default_state']
        )

        # update visibility of interval, time, and yscale radio buttons
        raw_classes = dashboard_inputs.compute_input_classes(
            display_state, dashboard.spec
        )
        classes = list(raw_classes.values())

        if ui_state['radio_group_visibility']:
            radio_group_row_class = ''
        else:
            radio_group_row_class = 'hidden'

        return (
            list(display_state.values())
            + classes
            + [
                url,
                ui_state['initial_load'],
                ui_state['help_open'],
                ui_state['submetric_options'],
                ui_state['radio_group_visibility'],
                radio_group_row_class,
            ]
        )

    # add chart callback
    @app.callback(*update_chart_args)  # type: ignore
    def update_chart(*args) -> go.Figure:
        kwargs = dict(zip(inputs.keys(), args))
        kwargs['now'] = kwargs['now'][:10]
        state = dashboard_state.build_state(
            _fix=False, _ui_spec=dashboard.spec, **kwargs
        )
        return dashboard.create_chart(state)

    # add clientside callbacks
    dashboard_inputs._prevent_button_focus(app)

    # Add a new route
    @app.server.route('/metadata')  # type: ignore
    def mince_metadata() -> flask.Response:
        import flask

        return flask.jsonify(dashboard.get_metadata())

    # add interactive debugger
    if pdb:

        @app.server.errorhandler(Exception)  # type: ignore
        def custom_exception_handler(
            e: Exception,
        ) -> tuple[flask.Response, int]:
            import sys
            import flask

            try:
                import ipdb as pdb  # type: ignore
            except ImportError:
                import pdb
            exc_type, exc_value, exc_traceback = sys.exc_info()
            traceback_details = {
                'filename': exc_traceback.tb_frame.f_code.co_filename,  # type: ignore
                'lineno': exc_traceback.tb_lineno,  # type: ignore
                'name': exc_traceback.tb_frame.f_code.co_name,  # type: ignore
                'type': exc_type.__name__,  # type: ignore
                'message': str(e),
            }
            pdb.post_mortem(exc_traceback)
            return flask.jsonify(error=traceback_details), 500

    return app


def _compute_decorator_args(
    ui_spec: types.UiSpec,
) -> tuple[list[Input | Output | State], list[Input | Output | State]]:
    from dash import Input, Output, State

    dash_input_values = []
    dash_output_values = []
    dash_output_classes = []
    for name, input_spec in ui_spec['inputs'].items():
        if input_spec['type'] == 'button':
            dash_input_values.append(Input(name + '-radio', 'value'))
            dash_output_values.append(Output(name + '-radio', 'value'))
            dash_output_classes.append(Output(name + '-radio', 'className'))
        elif input_spec['type'] == 'date':
            dash_input_values.append(Input('date-picker', 'date'))
            dash_output_values.append(Output('date-picker', 'date'))
            dash_output_classes.append(Output('date-picker', 'className'))
        else:
            raise Exception('invalid type')

    # ui args
    update_ui_args = [
        # outputs
        dash_output_values,
        dash_output_classes,
        [
            Output('url', 'search'),
            Output('initial-load', 'data'),
            Output('help-modal', 'is_open'),
            Output('submetric-radio', 'options'),
            Output('radio-group-visibility', 'data'),
            Output('radio-group-row', 'className'),
        ],
        # inputs
        dash_input_values,
        [
            Input('url', 'search'),
            State('initial-load', 'data'),
            Input('keyboard', 'n_keydowns'),
            State('keyboard', 'keydown'),
            State('help-modal', 'is_open'),
            State('submetric-radio', 'options'),
            State('radio-group-visibility', 'data'),
        ],
    ]

    # chart args
    update_chart_args = [
        Output('main-chart', 'figure'),
        dash_input_values,
    ]

    return update_ui_args, update_chart_args


def get_default_assets_folder() -> str:
    import importlib.resources
    import os

    path = str(
        importlib.resources.path(
            'mince.resources.frontend.assets', 'favicon.ico'
        )
    )
    return os.path.dirname(path)



"""mince is a toolkit for slicing up data into bite-sized dashboards"""

__version__ = '0.2.4'

import typing

from .dashboards import Dashboard
from .ops import async_collect_dashboard_data, run_dashboard
from .ops.collection.collection_piecewise import PiecewiseCollector

if typing.TYPE_CHECKING:
    from .types import *



from __future__ import annotations

import typing

from mince import specs

if typing.TYPE_CHECKING:
    from mince import types


def get_df_metric(state: dict[str, str], ui_spec: types.UiSpec) -> str:
    if state['metric'] in ui_spec['submetrics']:
        return state['submetric']
    else:
        return state['metric']


def build_state(
    *,
    _ui_spec: types.UiSpec,
    _defaults: bool = False,
    _fix: bool = True,
    _validate: bool = True,
    **kwargs: typing.Any,
) -> dict[str, str]:
    import tooltime

    _replace_aliased_values(kwargs, ui_spec=_ui_spec)
    state = dict(**kwargs)

    if _defaults:
        for key, value in _ui_spec['default_state'].items():
            if key not in state:
                state[key] = value
        if 'now' not in state:
            state['now'] = tooltime.now()

    if _fix or _validate:
        if _fix:
            _fix_invalid_option_combinations(state, _ui_spec)
        elif _validate:
            invalid_combos = _find_invalid_option_combinations(state, _ui_spec)
            if invalid_combos is not None:
                raise Exception(
                    'invalid combination of state: ' + str(invalid_combos)
                )
        invalid_values = _find_invalid_option_values(state, _ui_spec)
        if len(invalid_values) > 0:
            raise Exception('invalid option values: ' + str(invalid_values))

    return state


def _replace_aliased_values(
    combo: dict[str, types.StateDatum | None],
    ui_spec: types.UiSpec,
) -> None:
    for key, value in combo.items():
        if value in ui_spec['inputs'][key]['aliases']:
            if isinstance(value, str):
                combo[key] = ui_spec['inputs'][key]['aliases'][value]


def _fix_invalid_option_combinations(
    state: dict[str, types.StateDatum | None],
    ui_spec: types.UiSpec,
    freeze_fields: list[str] | None = None,
    aliased: bool = False,
) -> None:
    while True:
        invalid = _find_invalid_option_combinations(state, ui_spec)
        if invalid is None:
            return
        for key in invalid.keys():
            if freeze_fields is not None and key not in freeze_fields:
                state[key] = ui_spec['default_state'][key]


def _fix_invalid_inputs(
    state: dict[str, types.StateDatum | None],
    raw_shortcut: dict[str, typing.Any],
    ui_spec: types.UiSpec,
) -> None:
    from dash import callback_context  # type: ignore

    # determine origin of trigger
    ctx = callback_context
    triggered_input = ctx.triggered[0]['prop_id'].split('.')[0]

    # get shortcut spec
    if raw_shortcut is not None:
        shortcut = ui_spec['shortcuts'][raw_shortcut['key']]
    else:
        shortcut = {}

    # fix time_window too small for sample_interval too small
    _ensure_time_window_exceeds_sample_interval(
        state, triggered_input, shortcut
    )

    # check invalid option combinations
    trigger = triggered_input.split('-radio')[0]

    _fix_invalid_option_combinations(
        state, freeze_fields=[trigger], aliased=True, ui_spec=ui_spec
    )


def _ensure_time_window_exceeds_sample_interval(
    values: dict[str, types.StateDatum | None],
    trigger: str,
    shortcut: types.ShortcutSpec,
) -> None:
    if values['time_window'] == '7d' and values['sample_interval'] == 'weekly':
        if trigger == 'time_window-radio':
            values['sample_interval'] = 'daily'
        elif trigger == 'sample_interval-radio':
            values['time_window'] = '30d'
        elif trigger == 'keyboard' and shortcut['field'] == 'sample_interval':
            values['time_window'] = '30d'
        else:
            values['sample_interval'] = 'daily'
    elif (
        values['time_window'] == '7d' and values['sample_interval'] == 'monthly'
    ):
        if trigger == 'time_window-radio':
            values['sample_interval'] = 'daily'
        elif trigger == 'sample_interval-radio':
            values['time_window'] = '365d'
        elif trigger == 'keyboard' and shortcut['field'] == 'sample_interval':
            values['time_window'] = '365d'
        else:
            values['sample_interval'] = 'daily'
    elif (
        values['time_window'] == '30d'
        and values['sample_interval'] == 'monthly'
    ):
        if trigger == 'time_window-radio':
            values['sample_interval'] = 'weekly'
        elif trigger == 'sample_interval-radio':
            values['time_window'] = '365d'
        elif trigger == 'keyboard' and shortcut['field'] == 'sample_interval':
            values['time_window'] = '365d'
        else:
            values['sample_interval'] = 'weekly'


def _find_invalid_option_combinations(
    state: dict[str, types.StateDatum | None],
    ui_spec: types.UiSpec,
    aliased: bool = False,
) -> dict[str, typing.Any] | None:
    if aliased:
        state = state.copy()
        _replace_aliased_values(state, ui_spec)
    for value in ui_spec['invalid_states']:
        if all(state[k] == v for k, v in value.items()):
            return value
    return None


def _find_invalid_option_values(
    state: dict[str, str],
    ui_spec: types.UiSpec,
) -> dict[str, typing.Any]:
    output = {}
    for key, value in state.items():
        if ui_spec['inputs'][key]['type'] == 'button':
            valid_values = [
                specs.alias_to_state(option, key, ui_spec)
                for option in ui_spec['inputs'][key]['button_options']
            ]
        elif ui_spec['inputs'][key]['type'] == 'date':
            continue
        else:
            raise Exception('invalid type')

        if value not in valid_values:
            print('VALID VALUES:', valid_values, 'ACTUAL:', value, 'FOR:', key)
            output[key] = value
    return output



from __future__ import annotations

import typing
import importlib
import argparse

import mince
from . import cli_helpers


def parse_args(
    subcommands: list[str],
    include_dashboard_arg: bool,
    DashboardClass: typing.Type[mince.Dashboard] | None,
) -> argparse.Namespace:
    if DashboardClass is None:
        try:
            DashboardClass = parse_dashboard()
        except (SystemExit, KeyError):
            DashboardClass = None

    parser = argparse.ArgumentParser(
        formatter_class=cli_helpers.HelpFormatter, allow_abbrev=False
    )

    # add commands
    subparsers = parser.add_subparsers(dest='command')
    for command in subcommands:
        module_name = 'mince.cli.subcommands.' + command + '_command'
        module = importlib.import_module(module_name)
        arg_adder = getattr(module, 'add_' + command + '_command')
        arg_adder(
            subparsers,
            include_dashboard_arg=include_dashboard_arg,
            DashboardClass=DashboardClass,
        )

    # parse args
    args = parser.parse_args()

    if args.command is None:
        import sys

        parser.print_help()
        sys.exit(0)

    return args


def parse_dashboard() -> typing.Type[mince.Dashboard] | None:
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument('command')
    parser.add_argument('dashboard', nargs='?')
    add_common_args(parser)
    args, extras = parser.parse_known_args()
    dashboard_name = args.dashboard
    if dashboard_name is not None:
        registry: mince.RegistryFile = {
            'path': args.registry,
            'validate': not args.no_validate,
        }
        DashboardClass = mince.dashboards.get_dashboard_class(
            name=dashboard_name, registry=registry
        )
        return DashboardClass
    else:
        return None


def add_common_args(
    parser: argparse.ArgumentParser, registry_name: str = '--registry'
) -> None:
    parser.add_argument(
        registry_name,
        metavar='PATH',
        help='path to mince registry',
    )
    parser.add_argument(
        '--no-validate',
        help='skip validation of registry',
        action='store_true',
    )
    parser.add_argument(
        '--debug',
        help='use debug mode',
        action='store_true',
    )
    parser.add_argument(
        '--pdb',
        help='use interactive debugger',
        action='store_true',
    )
    parser.add_argument(
        '--cd-destination-tempfile',
        help=argparse.SUPPRESS,
    )



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    import polars as pl
    import plotly.graph_objects as go  # type: ignore


def plot_line(
    df: pl.DataFrame,
    metric: str,
    grouping: str,
    x: str = 'x',
    log_y: bool = False,
    relative_y: bool = False,
    xlabel: str | None = None,
    ylabel: str | None = None,
    ytick_format: typing.Literal['$', '%', 'bytes'] | None = None,
    title: str | None = None,
    hover: bool = True,
    total: bool = False,
    colors: typing.Mapping[str, str] | None = None,
) -> go.Figure:
    import plotly.express as px  # type: ignore
    import polars as pl
    import toolstr

    # create formatted values
    if ytick_format == '$':
        toolstr_kwargs: typing.Mapping[str, typing.Any] = {
            'order_of_magnitude': True,
            'decimals': 1,
            'prefix': '$',
        }
    elif ytick_format == '%':
        toolstr_kwargs = {'percentage': True, 'decimals': 1}
    elif ytick_format == 'bytes':
        max_value: float = df['value'].max()  # type: ignore
        if max_value < 1024:
            unit = 1
            unit_label = 'Bytes'
        elif max_value < 1024**2:
            unit = 1024
            unit_label = 'KB'
        elif max_value < 1024**3:
            unit = 1024**2
            unit_label = 'MB'
        elif max_value < 1024**4:
            unit = 1024**3
            unit_label = 'GB'
        else:
            unit = 1024**4
            unit_label = 'TB'
        toolstr_kwargs = {'postfix': ' ' + unit_label, 'decimals': 2}
        if ylabel is None:
            ylabel = ''
        ylabel = ylabel + ' (' + unit_label + ')'
        df = df.with_columns(pl.col.value / unit)
    else:
        toolstr_kwargs = {'order_of_magnitude': True, 'decimals': 1}
    df = df.with_columns(
        formatted_value=pl.col.value.map_elements(
            lambda x: toolstr.format(x, **toolstr_kwargs),
            return_dtype=pl.String,
        )
    )

    fig = px.line(
        df,
        x=x,
        y='value',
        color=grouping,
        color_discrete_map=colors,
        custom_data=['formatted_value'],
    )
    fig.update_traces(line=dict(width=3))

    if total:
        fig.update_traces(
            line=dict(width=5),
            selector=dict(name='TOTAL'),
        )

    label_font = {
        'size': 18,
        'color': 'black',
        'family': 'monospace',
    }
    if title is None:
        title = metric
    title_kwargs = {
        'text': title,
        'y': 0.99,
        'font': {'size': 28, 'color': 'black', 'family': 'monospace'},
        'x': 0.5,
        'xanchor': 'center',
        'yanchor': 'top',
    }
    axis = {
        'showgrid': True,
        'gridcolor': '#DFDFDF',
        'gridwidth': 1,
        'griddash': 'dot',
    }
    if ylabel is None:
        ylabel = metric
    fig.update_layout(
        hoverlabel={
            'font_size': 12,
        },
    )
    fig.update_layout(
        paper_bgcolor='white',
        plot_bgcolor='white',
        dragmode='zoom',
        xaxis=dict(axis, title=xlabel),
        yaxis=dict(axis, title=ylabel),
        bargap=0,
        bargroupgap=0,
        xaxis_fixedrange=True,
        xaxis_type='category',
        yaxis_fixedrange=True,
        legend_title=None,
        legend_font={'size': 18, 'color': 'black', 'family': 'monospace'},
        # legend_traceorder='reversed',
        title=title_kwargs,
        hovermode='x unified',
    )
    if log_y:
        fig.update_layout(yaxis_type='log')
    fig.update_traces(
        marker_line_width=0.1,
        marker_line_color='black',
        hoverinfo='text+x+y',
    )

    label_font = {
        'size': 18,
        'color': 'black',
        'family': 'monospace',
    }

    if df[x].dtype == pl.Datetime:
        fig.update_xaxes(
            nticks=7,
            # autotickangles=[90],
        )
        fig.update_layout(
            xaxis=dict(
                tickformat='%Y-%m-%d',
                type='date',
            ),
        )
    fig.update_xaxes(
        title_font=label_font,
        tickfont=label_font,
        showspikes=True,
        spikesnap='cursor',
        spikemode='across',
        spikethickness=1,
        spikecolor='black',
        spikedash='solid',
    )
    fig.update_yaxes(
        title_font=label_font,
        tickfont=label_font,
        showspikes=True,
        spikesnap='cursor',
        spikemode='across',
        spikethickness=1,
        spikecolor='black',
        spikedash='solid',
    )

    if hover:
        fig.update_traces(
            hovertemplate='%{fullData.name} %{customdata[0]}<extra></extra>'
        )
    else:
        fig.update_traces(hovertemplate=None)
        fig.update_traces(hoverinfo='none')

    # set ytick format
    if ytick_format is not None:
        if ytick_format == '$':
            fig.update_yaxes(tickprefix='$')
        elif ytick_format == '%':
            fig.update_yaxes(tickformat='.0%')
        elif ytick_format == 'bytes':
            pass
        else:
            raise Exception('invalid yformat: ' + str(ytick_format))

    return fig



from __future__ import annotations

from typing import TYPE_CHECKING

from . import schema_structure

if TYPE_CHECKING:
    from mince import types


def partial_schema_to_whole(
    partial: types.DataSchemaPartial,
) -> types.DataSchema:
    import polars as pl

    raw_group_columns = partial.get('group_columns')
    if raw_group_columns is None:
        group_columns = []
    else:
        group_columns = raw_group_columns

    columns = {}
    raw_columns = partial.get('columns')
    if raw_columns is None:
        raise Exception('must specify columns')
    elif isinstance(raw_columns, list):
        default_columns = schema_structure.get_default_columns()
        for column in raw_columns:
            dtype = default_columns.get(column)
            if dtype is not None:
                columns[column] = dtype
            else:
                columns[column] = pl.String
    elif isinstance(raw_columns, dict):
        default_columns = schema_structure.get_default_columns()
        for column, dtype in raw_columns.items():
            if dtype is not None:
                columns[column] = dtype
            elif column in default_columns:
                columns[column] = default_columns[column]
            else:
                columns[column] = pl.String
    else:
        raise Exception('invalid column format')

    return {
        'columns': columns,
        'group_columns': group_columns,
    }



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    import datetime
    import polars as pl

    class _PlotKwargs(typing.TypedDict, total=False):
        changes: pl.DataFrame
        interval: str
        metric: str
        group_by: list[str]
        min_relative_change: float | None
        min_abs_change: float | None
        metric_format: dict[str, typing.Any] | None


def compute_changes(
    df: pl.DataFrame,
    *,
    group_by: str | list[str],
    metric: str,
    intervals: list[str] | dict[str, pl.Expr] | None = None,
    now: datetime.datetime | None = None,
    timestamp_column: str = 'timestamp',
    allow_null_rows: bool = False,
    pre_aggregate: str | None = 'sum',
) -> pl.DataFrame:
    import polars as pl
    import datetime

    if pre_aggregate is not None:
        if isinstance(group_by, str):
            group_by = [group_by]
        if pre_aggregate == 'sum':
            df = df.group_by(timestamp_column, *group_by).agg(pl.sum(metric))
        else:
            raise Exception('invalid pre aggregation: ' + str(pre_aggregate))

    # determine now
    if now is None:
        now = (
            df[[timestamp_column, metric]]  # type: ignore
            .filter(pl.col(metric).is_not_null())[timestamp_column]
            .max()
        )
    if isinstance(now, datetime.date):
        now = datetime.datetime(now.year, now.month, now.day)
    if not isinstance(now, datetime.datetime):
        raise Exception('could not determine timestamp')

    # gather intervals
    if intervals is None:
        intervals = ['7d', '1mo', '1q', '1y']
    intervals = {
        interval: pl.lit(now).dt.offset_by('-' + interval)
        for interval in intervals
    }

    # build columns
    value_columns = {
        'value_' + interval: pl.col(metric)
        .filter(pl.col(timestamp_column) == timestamp)
        .last()
        for interval, timestamp in intervals.items()
    }
    abs_delta_columns = {
        'abs_delta_' + interval: pl.col.value_now - pl.col('value_' + interval)
        for interval in intervals.keys()
    }
    rel_delta_columns = {
        'rel_delta_' + interval: (
            pl.col.value_now - pl.col('value_' + interval)
        )
        / pl.col('value_' + interval)
        for interval in intervals.keys()
    }

    # compute changes
    changes = (
        df.group_by(group_by)
        .agg(
            value_now=pl.col(metric)
            .filter(pl.col(timestamp_column) == now)
            .last(),
            **value_columns,
        )
        .with_columns(
            **abs_delta_columns,
            **rel_delta_columns,
            t_now=pl.lit(now),
            **{'t_' + k: v for k, v in intervals.items()},
        )
    )

    # check for nulls
    if (
        not allow_null_rows
        and not changes.filter(pl.col.value_now.is_null()).is_empty()
    ):
        raise Exception(
            'some rows have missing values for timestamp, use allow_null_rows=True'
        )

    return changes


def summarize_changes(
    changes: pl.DataFrame,
    metric: str,
    interval: str,
    *,
    n: int = 10,
    metric_format: dict[str, typing.Any] | None = None,
    min_new_value: float | str | None = None,
    min_old_value: float | str | None = None,
) -> None:
    import polars as pl
    import toolstr

    if metric_format is None:
        metric_format = {}

    # filter data
    abs_column = 'abs_delta_' + interval
    rel_column = 'rel_delta_' + interval
    changes = changes.filter(
        pl.col(rel_column).is_not_null(), pl.col(rel_column).is_not_nan()
    )

    # new_value = 'value'
    t_start = changes['t_' + interval].last()
    t_now = changes['t_now'].last()
    new_value = 'value on\n' + str(t_now)[:10]
    display_abs = 'Δ ' + interval
    display_rel = '%Δ ' + interval
    old_value = 'value on\n' + str(t_start)[:10]

    # create table subsets
    group_by = changes.columns[: changes.columns.index('value_now')]
    selection = {
        old_value: 'value_' + interval,
        new_value: 'value_now',
        display_abs: abs_column,
        display_rel: rel_column,
    }
    display_changes = changes.select(*group_by, **selection)
    gainers = display_changes.filter(pl.col(display_abs) > 0)
    losers = display_changes.filter(pl.col(display_abs) < 0)
    abs_gainers = gainers.sort(display_abs, descending=True, nulls_last=True)
    abs_losers = losers.sort(display_abs, nulls_last=True)
    rel_gainers = gainers.sort(display_rel, descending=True, nulls_last=True)
    rel_losers = losers.sort(display_rel, nulls_last=True)
    if min_old_value is not None:
        rel_losers = rel_losers.filter(pl.col(new_value) >= min_old_value)
        abs_losers = abs_losers.filter(pl.col(new_value) >= min_old_value)
    if min_new_value is not None:
        rel_gainers = rel_gainers.filter(pl.col(new_value) >= min_new_value)
        abs_gainers = abs_gainers.filter(pl.col(new_value) >= min_new_value)

    # table formatting
    standard_format = {'order_of_magnitude': True, 'decimals': 1}
    column_formats: dict[str, dict[str, typing.Any]] = {
        new_value: standard_format,
        old_value: standard_format,
        display_abs: dict(standard_format, signed=True),
        display_rel: {'percentage': True, 'decimals': 1, 'signed': True},
    }
    column_formats[new_value].update(metric_format)
    column_formats[display_abs].update(metric_format)
    table_kwargs = {'column_formats': column_formats, 'indent': 4, 'compact': 4}

    # print preamble
    time_range = '(' + str(t_start)[:10] + ' to ' + str(t_now)[:10] + ')'
    toolstr.print_text_box(
        'Biggest movers for ' + metric + ', last ' + interval
    )

    # print largest gainers and losers
    print()
    toolstr.print_header(
        'Largest absolute gainers for `' + metric + '` ' + time_range
    )
    print('- n_entries:', toolstr.format(len(abs_gainers)))
    if min_new_value is not None:
        print('- filter: new_value > ', toolstr.format(min_new_value))
    print()
    toolstr.print_dataframe_as_table(abs_gainers[:n], **table_kwargs)  # type: ignore
    print()
    toolstr.print()
    print()
    toolstr.print_header('Largest relative gainers ' + time_range)
    print('- n_entries:', toolstr.format(len(rel_gainers)))
    if min_new_value is not None:
        print('- filter: new_value > ', toolstr.format(min_new_value))
    print()
    toolstr.print_dataframe_as_table(rel_gainers[:n], **table_kwargs)  # type: ignore
    print()
    toolstr.print()
    print()
    toolstr.print_header('Largest absolute losers ' + time_range)
    print('- n_entries:', toolstr.format(len(abs_losers)))
    if min_old_value is not None:
        print('- filter: old_value > ', toolstr.format(min_old_value))
    print()
    toolstr.print_dataframe_as_table(abs_losers[:n], **table_kwargs)  # type: ignore
    print()
    toolstr.print()
    print()
    toolstr.print_header('Largest relative losers ' + time_range)
    print('- n_entries:', toolstr.format(len(rel_losers)))
    if min_old_value is not None:
        print('- filter: old_value > ', toolstr.format(min_old_value))
    print()
    toolstr.print_dataframe_as_table(rel_losers[:n], **table_kwargs)  # type: ignore

    # print entries
    print()
    toolstr.print()
    print()
    entries = changes.sort('value_now', descending=True, nulls_last=True)
    entries = entries.select(*group_by, **selection)
    toolstr.print_header('Largest current entries')
    print('- n_entries:', toolstr.format(len(entries)))
    toolstr.print_dataframe_as_table(entries[:n], **table_kwargs)  # type: ignore


def _get_intervals(changes: pl.DataFrame) -> list[str]:
    return [
        column.split('abs_delta_')[1]
        for column in changes.columns
        if column.startswith('abs_delta_')
    ]


def _get_group_by(changes: pl.DataFrame) -> list[str]:
    return changes.columns[: changes.columns.index('value_now')]


def plot_changes_vs_size(
    changes: pl.DataFrame,
    metric: str,
    *,
    intervals: list[str] | None = None,
    group_by: list[str] | None = None,
    plot_kwargs: _PlotKwargs | None = None,
    min_relative_change: float | None = None,
    min_abs_change: float | None = None,
    metric_format: dict[str, typing.Any] | None = None,
) -> None:
    import toolplot

    toolplot.setup_plot_formatting()

    if intervals is None:
        intervals = _get_intervals(changes)
    if group_by is None:
        group_by = _get_group_by(changes)
    kwargs: _PlotKwargs = dict(
        changes=changes,
        group_by=group_by,
        metric=metric,
        min_relative_change=min_relative_change,
        min_abs_change=min_abs_change,
        metric_format=metric_format,
    )
    if plot_kwargs is not None:
        kwargs.update(plot_kwargs)
    for interval in intervals:
        kwargs['interval'] = interval
        _plot_changes(positive=True, change_type='relative', **kwargs)
        _plot_changes(positive=False, change_type='relative', **kwargs)
        _plot_changes(positive=True, change_type='absolute', **kwargs)
        _plot_changes(positive=False, change_type='absolute', **kwargs)


def _plot_changes(
    changes: pl.DataFrame,
    interval: str,
    metric: str,
    group_by: list[str],
    change_type: typing.Literal['absolute', 'relative'],
    positive: bool = True,
    min_relative_change: float | None = None,
    min_abs_change: float | None = None,
    metric_format: dict[str, typing.Any] | None = None,
) -> None:
    import polars as pl
    import matplotlib.pyplot as plt
    import toolplot

    # select subset of data
    if min_abs_change is None:
        min_abs_change = 0
    if min_relative_change is None:
        min_relative_change = 0.01
    if positive:
        data = changes.filter(
            pl.col('abs_delta_' + interval) > min_abs_change,
            pl.col.value_now > 0,
        )
        if change_type == 'relative':
            data = data.filter(
                pl.col('rel_delta_' + interval) > min_relative_change
            )
            yvals = data['rel_delta_' + interval]
        else:
            yvals = data['abs_delta_' + interval]
        xvals = data['value_now']
        color = 'green'
    else:
        data = changes.filter(
            pl.col('abs_delta_' + interval) < -min_abs_change,
            pl.col.value_now > 0,
        )
        if change_type == 'relative':
            data = data.filter(
                pl.col('rel_delta_' + interval) < -min_relative_change
            )
            yvals = -data['rel_delta_' + interval]
        else:
            yvals = -data['abs_delta_' + interval]
        xvals = data['value_' + interval]
        color = 'red'

    # skip if empty
    if len(data) == 0:
        print('no data, skipping plot')
        return

    # create scatterplot
    plt.plot(xvals, yvals, '.', color=color, markersize=10)
    name: list[str] | tuple[str] | None
    for name, x, y in zip(data[group_by].rows(), xvals, yvals):
        if name is None or name[0] is None:
            name = ['unlabeled'] * len(group_by)
        if x is not None and y is not None:
            plt.text(
                x,
                y,
                ' ' + '_'.join(name),
                fontsize=8,
            )
    plt.xscale('log')

    plt.yscale('log')
    if change_type == 'absolute' and min_abs_change > 0:
        ylim = plt.ylim()
        plt.ylim([max(ylim[0], min_abs_change), ylim[1]])

    if positive:
        plt.xlabel('\ncurrent value')
        if change_type == 'relative':
            toolplot.format_yticks(toolstr_kwargs={'percentage': True})
        else:
            ytick_format: dict[str, typing.Any] = {
                'order_of_magnitude': True,
                'decimals': 0,
            }
            if metric_format is not None:
                ytick_format.update(metric_format)
            toolplot.format_yticks(toolstr_kwargs=ytick_format)
    else:
        plt.xlabel('\nvalue ' + interval + ' ago')
        if change_type == 'relative':
            toolplot.format_yticks(
                toolstr_kwargs={'percentage': True, 'prefix': '–'}
            )
            plt.ylim([1, min_relative_change])
        else:
            ytick_format = {
                'order_of_magnitude': True,
                'decimals': 0,
                'prefix': '',
            }
            if metric_format is not None:
                ytick_format.update(metric_format)
            ytick_format['prefix'] = '−' + ytick_format['prefix']
            toolplot.format_yticks(toolstr_kwargs=ytick_format)
            ylim = plt.ylim()
            plt.ylim([ylim[1], ylim[0]])

    # label plot
    xtick_format = {'order_of_magnitude': True, 'decimals': 1}
    if metric_format is not None:
        xtick_format.update(metric_format)
    toolplot.format_xticks(
        toolstr_kwargs=xtick_format,
        rotation=0,
    )
    if change_type == 'relative':
        plt.ylabel('% change over ' + interval + '\n')
    else:
        plt.ylabel('abs change over ' + interval + '\n')
    toolplot.add_tick_grid()
    t_start = changes['t_' + interval].last()
    t_now = changes['t_now'].last()
    time_range = '(' + str(t_start)[:10] + ' to ' + str(t_now)[:10] + ')'
    plt.title(
        change_type.title()
        + ' changes in '
        + metric
        + ' for each '
        + '+'.join(group_by)
        + '\nduration = '
        + interval
        + ' '
        + time_range
    )
    plt.show()  # type: ignore



from __future__ import annotations

import typing

from . import spec_validate

if typing.TYPE_CHECKING:
    from typing import Mapping, Sequence, Union
    from mince.types import (
        StateDatum,
        InputSpec,
        UiSpec,
        ShortcutSpec,
        MetricSpec,
        DataSchema,
    )


def create_ui_spec(
    *,
    name: str,
    title: str | None,
    description: str | None,
    favicon_emoji: str | None = None,
    version: str,
    invalid_states: Sequence[Mapping[str, StateDatum]],
    inputs: Mapping[str, InputSpec],
    metrics: dict[str, MetricSpec],
    groupings: list[str],
    submetrics: Mapping[str, Sequence[str]],
    shortcuts: Mapping[str, ShortcutSpec],
    colors: Mapping[str, str],
    schema: DataSchema,
    title_prefix: str | None,
    title_infix: str | None,
    title_postfix: str | None,
) -> UiSpec:
    default_state: dict[str, Union[str, int, float] | None] = {
        input: input_spec['default'] for input, input_spec in inputs.items()
    }

    ui_spec: UiSpec = {
        'name': name,
        'title': title,
        'description': description,
        'favicon_emoji': favicon_emoji,
        'version': version,
        'groupings': groupings,
        'default_state': default_state,
        'invalid_states': [dict(item) for item in invalid_states],
        'inputs': dict(inputs),
        'metrics': metrics,
        'submetrics': {k: list(v) for k, v in submetrics.items()},
        'shortcuts': dict(shortcuts),
        'colors': dict(colors),
        'schema': schema,
        'title_prefix': title_prefix,
        'title_infix': title_infix,
        'title_postfix': title_postfix,
    }

    spec_validate.validate_ui_spec(ui_spec)

    return ui_spec



from __future__ import annotations

import typing

if typing.TYPE_CHECKING:
    import polars as pl
    import plotly.graph_objects as go  # type: ignore


def create_treemap(
    df: pl.DataFrame,
    *,
    title: str,
    metric: str,
    grouping: str,
    subgrouping: str | None = None,
    prefix: str | None = None,
    colors: typing.Mapping[str, str] | None = None,
) -> go.Figure:
    import polars as pl
    import toolstr
    import plotly.graph_objects as go

    toolstr_kwargs: dict[str, typing.Any] = {
        'order_of_magnitude': True,
        'decimals': 1,
    }
    if prefix is not None:
        toolstr_kwargs['prefix'] = prefix

    grand_total = df['value'].sum()

    names = []
    sizes = []
    parents: list[str | None] = []
    custom_data = []

    names.append(title)
    sizes.append(grand_total)
    parents.append(None)
    custom_data.append(
        '<b>{name}</b><br>'.format(name=title)
        + toolstr.format(grand_total, **toolstr_kwargs)
        + ' total '
        + metric
    )
    top_parent = title

    grouped = (
        df.group_by(grouping)
        .agg(pl.sum('value'))
        .sort('value', descending=True)
    )
    for group, value in grouped.rows():
        names.append(group)
        sizes.append(value)
        parents.append(top_parent)
        custom_data.append(
            '<b>{name}</b><br>'.format(name=group)
            + toolstr.format(value, **toolstr_kwargs)
            + ' '
            + metric
            + '<br>'
            + toolstr.format(value / grand_total, percentage=True, decimals=1)
            + ' of total'
        )

    if subgrouping is not None:
        for group, value in grouped.rows():
            other_grouped = df.filter(pl.col(grouping) == group)
            other_total = other_grouped['value'].sum()
            other_grouped = other_grouped.with_columns(
                prop=(pl.col.value / pl.col.value.sum()),
            )  # .filter(pl.col.prop > 0.001)
            if len(other_grouped) >= 2:
                for other_group, value in other_grouped[
                    [subgrouping, 'value']
                ].rows():
                    while other_group in names:
                        other_group = ' ' + other_group + ' '
                    names.append(other_group)
                    sizes.append(value)
                    parents.append(group)
                    if grand_total == 0:
                        grand_fraction = 0
                    else:
                        grand_fraction = value / grand_total
                    if other_total == 0:
                        other_fraction = 0
                    else:
                        other_fraction = value / other_total
                    custom_data.append(
                        (
                            '<b>{name}</b><br>'.format(
                                name=other_group + ' ' + group
                            )
                            + toolstr.format(value, **toolstr_kwargs)
                            + ' '
                            + metric
                            + '<br>'
                            + toolstr.format(
                                grand_fraction, percentage=True, decimals=1
                            )
                            + ' of total'
                            + '<br>'
                            + toolstr.format(
                                other_fraction, percentage=True, decimals=1
                            )
                            + '% of {group}'.format(group=group)
                        )
                    )

    fig = go.Figure(
        go.Treemap(
            labels=names,
            parents=parents,
            values=sizes,
            textfont=dict(
                family='monospace',
                size=28,
            ),
            branchvalues='total',
            customdata=custom_data,
        ),
    )

    fig.update_traces(
        textposition='middle center',
        hovertemplate='%{customdata}<extra></extra>',
        hoverlabel={'font': {'size': 22, 'family': 'Monospace'}},
    )

    fig.update_traces(selector=dict(label=title), textfont=dict(color='red'))

    margin = 1
    fig.update_layout(
        margin=dict(t=margin, l=margin, r=margin, b=margin),
    )
    if colors is not None:
        fig.update_layout(
            treemapcolorway=[
                colors.get(group)
                for group in grouped[grouping]
                if group in colors
            ],
        )

    return fig



from __future__ import annotations

import os
import typing

from mince import dashboards

if typing.TYPE_CHECKING:
    from typing import Type, TypeVar
    from typing_extensions import TypeGuard
    from mince.types import (
        Registry,
        RegistryReference,
        RegistryEntry,
        CollectKwargsPartial,
    )

    T = TypeVar('T', bound=dashboards.Dashboard)


#
# # data loading
#


def get_registry_entry(
    name: str, registry: RegistryReference = None
) -> RegistryEntry:
    """load entry from registry"""
    registry = resolve_registry(registry)
    return registry['dashboards'][name]


#
# # reference resolution
#


def resolve_registry(registry: RegistryReference = None) -> Registry:
    """return Registry when reference is given as path or default (None)"""
    if isinstance(registry, str):
        registry = {'path': registry}
    if registry is None:
        return load_registry_file()
    elif isinstance(registry, dict) and 'path' in registry:
        return load_registry_file(
            path=registry.get('path'),  # type: ignore
            validate=registry.get('validate', True),  # type: ignore
            create_if_dne=registry.get('create_if_dne', True),  # type: ignore
        )
    elif isinstance(registry, dict) and 'dashboards' in registry:
        return registry  # type: ignore
    else:
        raise Exception('invalid registry reference')


#
# # registration
#


def register_dashboard(
    dashboard: str | Type[dashboards.Dashboard],
    *,
    registry: RegistryReference | None = None,
    collect_kwargs: CollectKwargsPartial | None = None,
) -> None:
    # resolve Dashboard class
    if isinstance(dashboard, str):
        if '.' in dashboard:
            DashboardClass = dashboards.resolve_dashboard_class(dashboard)
            dashboard_reference = dashboard
        else:
            DashboardClass = dashboards.find_dashboard_class(dashboard)
            dashboard_reference = (
                DashboardClass.__module__ + '.' + DashboardClass.__name__
            )
    elif issubclass(dashboard, dashboards.Dashboard):
        DashboardClass = dashboard
        dashboard_reference = (
            DashboardClass.__module__ + '.' + DashboardClass.__name__
        )
    else:
        raise Exception('dashboard class is not a subclass of Dashboard')

    # get metadata
    spec = DashboardClass.load_spec()
    name = spec['name']
    description = spec['description']
    if collect_kwargs is None:
        collect_kwargs = {}

    # add entry to registry
    entry: RegistryEntry = {
        'name': name,
        'dashboard_class': dashboard_reference,
        'description': description,
        'collect_kwargs': collect_kwargs,
    }
    resolved = resolve_registry(registry)
    resolved['dashboards'][name] = entry

    # save registry to disk
    if registry is None:
        save_registry_file(resolved, path=None)
    elif 'path' in registry:
        save_registry_file(
            resolved,
            path=registry['path'],  # type: ignore
        )

    print('registered', name)


def unregister_dashboard(
    name: str | typing.Type[dashboards.Dashboard],
    *,
    idempotent: bool = False,
    path: str | None = None,
    validate: bool = True,
) -> None:
    registry = load_registry_file(path=path, validate=validate)
    if isinstance(name, str):
        if name not in registry['dashboards'] and not idempotent:
            raise Exception('dashboard is not registered')
        del registry['dashboards'][name]
    elif type(name) is type and issubclass(name, dashboards.Dashboard):
        dashboard_path = name.__module__ + '.' + name.__name__
        for candidate_name, candidate in registry['dashboards'].items():
            if candidate['dashboard_class'] == dashboard_path:
                del registry['dashboards'][candidate_name]
    else:
        raise Exception('invalid type: ' + str(name))
    save_registry_file(registry=registry, path=path)


#
# # file io
#


def create_blank_registry() -> Registry:
    import mince

    return {
        'mince_version': mince.__version__,
        'dashboards': {},
    }


def load_registry_file(
    *,
    path: str | None = None,
    validate: bool = True,
    create_if_dne: bool = True,
) -> Registry:
    import json
    from mince import storage

    if path is None:
        path = storage.get_registry_path()
    if not os.path.exists(path):
        if create_if_dne:
            registry = create_blank_registry()
            save_registry_file(registry, path=path)
        else:
            raise Exception('registry file does not exist')
    else:
        with open(path, 'r') as f:
            registry = json.load(f)
    if validate_registry(registry):
        return registry
    else:
        raise Exception('invalid registry')


def save_registry_file(registry: Registry, *, path: str | None) -> None:
    import json
    from mince import storage

    if path is None:
        path = storage.get_registry_path()

    validate_registry(registry)

    os.makedirs(os.path.dirname(path), exist_ok=True)
    with open(path, 'w') as f:
        json.dump(registry, f)


#
# # validation
#


def validate_registry(registry: typing.Any) -> TypeGuard[Registry]:
    if not isinstance(registry, dict):
        raise Exception('invalid type for registry: ' + str(type(registry)))
    assert set(registry.keys()) == {'mince_version', 'dashboards'}
    assert isinstance(registry['mince_version'], str)
    assert isinstance(registry['dashboards'], dict)
    for key, value in registry['dashboards'].items():
        assert isinstance(key, str)
        validate_registry_entry(value)
    return True


def validate_registry_entry(entry: typing.Any) -> TypeGuard[RegistryEntry]:
    assert isinstance(entry, dict)
    assert set(entry.keys()) == {
        'name',
        'dashboard_class',
        'description',
        'collect_kwargs',
    }
    assert isinstance(entry['name'], str)
    assert isinstance(entry['dashboard_class'], str)
    assert entry['description'] is None or isinstance(entry['description'], str)
    assert isinstance(entry['collect_kwargs'], dict)
    return True



from __future__ import annotations

import argparse
import typing

import mince.storage
from .. import cli_helpers
from .. import cli_parsing

if typing.TYPE_CHECKING:
    from mince import types


shell_snippet = """function mince {
    local tempfile="$(mktemp -t tmp.XXXXXX)"
    if [[ -z "$@" ]]; then
        command mince
    else
        command mince "$@" --cd-destination-tempfile "$tempfile"
    fi
    if [[ -s "$tempfile" ]]; then
        cd "$(realpath $(cat "$tempfile"))"
    fi
    rm -f "$tempfile" 2>/dev/null
}"""


def add_cd_command(
    subparsers: cli_helpers.Subparsers,
    include_dashboard_arg: bool,
    DashboardClass: typing.Type[mince.Dashboard] | None,
) -> None:
    parser = subparsers.add_parser(
        'cd',
        help='cd to mince root or dashboard data directory',
        formatter_class=cli_helpers.HelpFormatter,
    )
    parser.set_defaults(f_command=cd_command)
    if include_dashboard_arg:
        parser.add_argument('dashboard', help='dashboard name', nargs='?')

    parser.add_argument(
        '--code', help='cd to code directory', action='store_true'
    )
    cli_parsing.add_common_args(parser)


def cd_command(
    args: argparse.Namespace,
    dashboard: str | typing.Type[types.Dashboard] | None = None,
) -> None:
    import os

    if args.cd_destination_tempfile is None:
        print(
            'to enable `mince cd`, add this to your shell config (e.g. ~/.bashrc):'
        )
        print()
        print(shell_snippet)
    else:
        path = get_path(args, dashboard)
        os.makedirs(path, exist_ok=True)
        with open(args.cd_destination_tempfile, 'w') as f:
            f.write(path)


def get_path(
    args: argparse.Namespace,
    dashboard: str | typing.Type[types.Dashboard] | None = None,
) -> str:
    import importlib
    import os

    registry: types.RegistryFile = {
        'path': args.registry,
        'validate': not args.no_validate,
    }
    if dashboard is None:
        dashboard = args.dashboard
    if isinstance(dashboard, str):
        DashboardClass = mince.dashboards.get_dashboard_class(
            dashboard, registry=registry
        )
    elif type(dashboard) is mince.Dashboard and issubclass(
        dashboard, mince.Dashboard
    ):
        DashboardClass = dashboard
    else:
        DashboardClass = None
    if DashboardClass is None:
        if args.code:
            return os.path.dirname(mince.__file__)
        else:
            return mince.storage.get_mince_root()
    else:
        if args.code:
            module = importlib.import_module(DashboardClass.__module__)
            return os.path.dirname(module.__file__)  # type: ignore
        else:
            return mince.storage.resolve_data_dir(
                dashboard=DashboardClass,
                registry=registry,
            )



from __future__ import annotations

import argparse
import typing

import mince
from ... import dashboards
from .. import cli_helpers
from .. import cli_parsing

if typing.TYPE_CHECKING:
    from mince import types


def add_spec_command(
    subparsers: cli_helpers.Subparsers,
    include_dashboard_arg: bool,
    DashboardClass: typing.Type[mince.Dashboard] | None,
) -> None:
    parser = subparsers.add_parser(
        'spec',
        help='view dashboard ui specification',
        formatter_class=cli_helpers.HelpFormatter,
    )
    parser.set_defaults(f_command=spec_command)
    if include_dashboard_arg:
        parser.add_argument(
            'dashboard',
            help='dashboard to get UI spec of',
        )
    parser.add_argument(
        '-i',
        '--interactive',
        help='load spec in interactive python session',
        action='store_true',
    )
    cli_parsing.add_common_args(parser)


def spec_command(
    args: argparse.Namespace,
    dashboard: str | typing.Type[types.Dashboard] | None = None,
) -> None:
    registry: types.RegistryFile = {
        'path': args.registry,
        'validate': not args.no_validate,
    }
    if dashboard is None:
        dashboard = args.dashboard
    if isinstance(dashboard, str):
        DashboardClass = dashboards.get_dashboard_class(
            dashboard, registry=registry
        )
    elif issubclass(dashboard, mince.Dashboard):
        DashboardClass = dashboard
    else:
        raise Exception()

    ui_spec = DashboardClass.load_spec()

    if args.interactive:
        cli_helpers.open_interactive_session(variables={'spec': ui_spec})
    else:
        import json

        as_str = json.dumps(
            ui_spec,
            sort_keys=True,
            indent=4,
            cls=cli_helpers.SpecJsonEncoder,
        )
        print(as_str)



from __future__ import annotations

import typing

import mince

if typing.TYPE_CHECKING:
    from . import piecewise_collector


def print_summary(
    output: piecewise_collector.Output,
    verbose: int,
    max_concurrent_queries: int,
) -> None:
    if verbose >= 2:
        import toolstr

        print()
        toolstr.print_header('Skipping chunks')
        if len(output['skip_slugs']) == 0:
            print('[none]')
        for slug in output['skip_slugs']:
            print(slug)
        print()
        toolstr.print_header('Running chunks')
        if len(output['run_slugs']) == 0:
            print('[none]')
        for slug in output['run_slugs']:
            print(slug)

    if verbose >= 1:
        n_run = len(output['coroutines'])
        n_skip = len(output['skip_slugs'])
        n_concurrent = min(n_run, max_concurrent_queries)
        print(
            '\n\nrunning',
            n_run,
            'collection chunks,',
            n_concurrent,
            'chunks at a time',
            '(skipping',
            n_skip,
            'existing chunks)',
            '\n',
        )


def get_chunk_slug(
    start_time: int | float,
    end_time: int | float,
    tags: dict[str, str],
    interval: mince.Interval,
    data_dir: str | None = None,
    path_template: str | None = None,
    verbose: int = 1,
    skip_incomplete_intervals: bool = False,
) -> str:
    import tooltime

    start_str = tooltime.timestamp_to_date(start_time)
    end_str = tooltime.timestamp_to_date(end_time)
    tag_str = [k + '=' + v for k, v in tags.items()]
    return ' '.join([start_str, end_str, 'interval=' + interval, *tag_str])



from __future__ import annotations

from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from mince import types
    import polars as pl


def validate_data(df: pl.DataFrame) -> None:
    """validate that df conforms to data schema"""
    import polars as pl

    valid_intervals = {'point', 'day', 'week', 'month', 'year'}
    invalid_intervals = df[['interval']].filter(
        ~pl.col.interval.is_in(valid_intervals)
    )
    if len(invalid_intervals) > 0:
        raise Exception('invalid value for interval: ' + str(invalid_intervals))


def validate_data_matches_spec(
    dfs: dict[str, pl.DataFrame], spec: types.UiSpec
) -> None:
    for df in dfs.values():
        validate_data(df)
        validate_schemas_equal(dict(df.schema), spec['schema']['columns'])


def validate_schemas_equal(
    data_schema: dict[str, Any], spec_schema: dict[str, Any]
) -> None:
    if data_schema == spec_schema:
        return None

    missing_columns = []
    for column in spec_schema:
        if column not in data_schema:
            missing_columns.append(column)
    if len(missing_columns) > 0:
        raise Exception(
            'data does not match schema, missing columns: '
            + str(missing_columns)
        )

    extra_columns = []
    for column in data_schema:
        if column not in spec_schema:
            extra_columns.append(column)
    if len(extra_columns) > 0:
        raise Exception(
            'data does not match schema, extra columns: ' + str(extra_columns)
        )

    different_types = {}
    for column in data_schema.keys():
        if data_schema[column] != spec_schema[column]:
            different_types[column] = (data_schema[column], spec_schema[column])
    if len(different_types) > 0:
        raise Exception(
            'data does not match schema, types do not match: '
            + str(different_types)
        )

    if list(data_schema.keys()) != list(spec_schema.keys()):
        raise Exception(
            'data does not match schema, columns are in wrong order'
        )

    raise Exception('data does not match schema, unknown difference')



from __future__ import annotations

import typing

import mince
from mince import Dashboard
from . import cli_parsing
from . import cli_helpers

if typing.TYPE_CHECKING:
    import argparse


def run_entire_cli() -> None:
    subcommands = [
        'ls',
        'cd',
        'dir',
        'register',
        'unregister',
        'validate',
        'collect',
        'run',
        'kill',
        'data',
        'spec',
        'docker',
    ]

    args = cli_parsing.parse_args(
        subcommands=subcommands,
        include_dashboard_arg=True,
        DashboardClass=None,
    )

    _execute(args, None)


def run_single_dashboard_cli(DashboardClass: type[Dashboard]) -> None:
    subcommands = [
        'cd',
        'dir',
        'register',
        'unregister',
        'validate',
        'collect',
        'run',
        'kill',
        'data',
        'spec',
    ]

    args = cli_parsing.parse_args(
        subcommands=subcommands,
        include_dashboard_arg=False,
        DashboardClass=DashboardClass,
    )

    _execute(args, DashboardClass)


def _execute(
    args: argparse.Namespace, DashboardClass: type[Dashboard] | None
) -> None:
    import inspect

    if args.pdb:
        try:
            if inspect.iscoroutinefunction(args.f_command):
                import asyncio

                asyncio.run(args.f_command(args, DashboardClass))
            else:
                args.f_command(args, DashboardClass)
        except Exception:
            cli_helpers._enter_debugger()
    else:
        if inspect.iscoroutinefunction(args.f_command):
            import asyncio

            asyncio.run(args.f_command(args, DashboardClass))
        else:
            args.f_command(args, DashboardClass)



from __future__ import annotations

import typing

import mince

from . import piecewise_io
from .piecewise_io import pieces_already_exist
from . import piecewise_summary
from . import piecewise_times

if typing.TYPE_CHECKING:
    T = typing.TypeVar('T')
    import polars as pl

    class Chunk(typing.TypedDict):
        start_time: float
        end_time: float
        interval: mince.Interval
        skip_incomplete_intervals: bool
        data_dir: str
        path_template: str
        tags: dict[str, str]
        verbose: int

    class Output(typing.TypedDict):
        coroutines: list[typing.Coroutine[typing.Any, typing.Any, None]]
        run_slugs: list[str | None]
        skip_slugs: list[str | None]


async def async_collect_data_piecewise(
    data_dir: str,
    path_template: str,
    start_time: float,
    end_time: float,
    intervals: list[mince.Interval],
    skip_incomplete_intervals: bool,
    max_concurrent_queries: int,
    chunk_size: str,
    skip_existing_files: bool,
    async_collect_df: mince.AsyncDfCollector,
    tags: dict[str, list[str]],
    get_chunk_time_range: typing.Callable[..., tuple[int, int]] | None = None,
    verbose: int = 1,
    dry: bool = False,
) -> dict[str, pl.DataFrame]:
    # collect pieces
    await _async_collect_data_pieces(
        data_dir=data_dir,
        start_time=start_time,
        end_time=end_time,
        intervals=intervals,
        skip_incomplete_intervals=skip_incomplete_intervals,
        max_concurrent_queries=max_concurrent_queries,
        chunk_size=chunk_size,
        skip_existing_files=skip_existing_files,
        async_collect_df=async_collect_df,
        get_chunk_time_range=get_chunk_time_range,
        path_template=path_template,
        tags=tags,
        verbose=verbose,
        dry=dry,
    )

    if dry:
        return {}

    # load results
    return piecewise_io.read_piece_files(
        data_dir=data_dir,
        path_template=path_template,
        tag_names=list(tags.keys()),
    )


async def _async_collect_data_pieces(
    data_dir: str,
    start_time: float,
    end_time: float,
    intervals: list[mince.Interval],
    skip_incomplete_intervals: bool,
    max_concurrent_queries: int,
    chunk_size: str,
    skip_existing_files: bool,
    async_collect_df: mince.AsyncDfCollector,
    get_chunk_time_range: typing.Callable[..., tuple[int, int]] | None,
    path_template: str,
    tags: dict[str, list[str]],
    verbose: int,
    dry: bool,
) -> None:
    if verbose >= 3:
        print('\nFile existence checks\n' + '─' * 21)
    output: Output = {'coroutines': [], 'run_slugs': [], 'skip_slugs': []}
    for tag_combo in compute_tag_combos(tags):
        for interval in intervals:
            chunks = piecewise_times.compute_time_chunks(
                start_time=start_time,
                end_time=end_time,
                chunk_size=chunk_size,
                interval=interval,
                tags=tag_combo,
                get_chunk_time_range=get_chunk_time_range,
                skip_incomplete_intervals=False,
            )
            for chunk_start, chunk_end in chunks[['start', 'end']].rows():
                chunk: Chunk = {
                    'start_time': chunk_start,
                    'end_time': chunk_end,
                    'interval': interval,
                    'skip_incomplete_intervals': skip_incomplete_intervals,
                    'data_dir': data_dir,
                    'path_template': path_template,
                    'tags': tag_combo,
                    'verbose': verbose,
                }
                slug = None
                if verbose >= 2:
                    slug = piecewise_summary.get_chunk_slug(**chunk)
                if skip_existing_files and pieces_already_exist(**chunk):
                    output['skip_slugs'].append(slug)
                    continue
                output['run_slugs'].append(slug)
                coroutine = _async_collect_piece_chunk(
                    dry=dry, async_collect_df=async_collect_df, **chunk
                )
                output['coroutines'].append(coroutine)

    piecewise_summary.print_summary(output, verbose, max_concurrent_queries)

    await semaphore_gather(max_concurrent_queries, *output['coroutines'])


async def _async_collect_piece_chunk(
    *,
    data_dir: str,
    path_template: str,
    interval: mince.Interval,
    skip_incomplete_intervals: bool,
    start_time: int | float,
    end_time: int | float,
    async_collect_df: mince.AsyncDfCollector,
    tags: dict[str, str],
    verbose: int,
    dry: bool,
) -> None:
    import polars as pl
    import tooltime

    # print summary
    if verbose >= 1:
        slug = piecewise_summary.get_chunk_slug(
            start_time=start_time,
            end_time=end_time,
            tags=tags,
            interval=interval,
        )
        print('doing', slug)

    if dry:
        return

    # get data
    df = await async_collect_df(
        interval=interval,
        start_time=tooltime.timestamp_to_seconds(start_time),
        end_time=tooltime.timestamp_to_seconds(end_time),
        tags=tags,
    )

    # add completeness information
    targets = tooltime.get_intervals(
        interval=interval,
        start=start_time,
        end=end_time,
        include_incomplete=not skip_incomplete_intervals,
    ).select(timestamp='start', complete=pl.col.completeness == 'complete')
    df = df.join(targets, on='timestamp', how='inner')
    if df['complete'].is_null().any():
        raise Exception('unexpected timestamps in collected data')

    # write pieces
    piecewise_io.write_piece_files(
        data_dir=data_dir,
        path_template=path_template,
        df=df,
        interval=interval,
        tags=tags,
    )


def compute_tag_combos(
    tags: dict[str, list[str]],
) -> typing.Generator[dict[str, str]]:
    import itertools

    for tag_combo_values in itertools.product(*tags.values()):
        yield dict(zip(tags.keys(), tag_combo_values))


async def semaphore_gather(
    n_concurrent: int, *tasks: typing.Awaitable[T]
) -> list[T]:
    import asyncio

    semaphore = asyncio.Semaphore(n_concurrent)

    async def semaphore_task(task: typing.Awaitable[T]) -> T:
        async with semaphore:
            return await task

    return await asyncio.gather(*(semaphore_task(task) for task in tasks))



from __future__ import annotations

import typing

from mince import transforms
from ... import raw_figures
from . import timeseries_data
from . import timeseries_styles

if typing.TYPE_CHECKING:
    import polars as pl
    import plotly.graph_objects as go  # type: ignore
    from mince import types


def create_time_series_fig(
    df: pl.DataFrame,
    state: dict[str, typing.Any],
    debug: bool,
    ui_spec: types.UiSpec,
    title: str | None = None,
) -> go.Figure:
    import time

    data_date_range = transforms.get_date_range(df)

    # prepare data
    start_time = time.time()
    df = timeseries_data._prepare_time_series_data(df, state, ui_spec=ui_spec)
    end_time = time.time()

    # create style kwargs
    style_kwargs = timeseries_styles._get_timeseries_style_kwargs(
        state, df, data_date_range=data_date_range, ui_spec=ui_spec, title=title
    )

    # print debug messages
    if debug:
        print()
        print('filtered data')
        print(
            'data filtering took',
            '%.6f' % ((end_time - start_time) * 1000),
            'milliseconds',
        )
        print('columns:', df.columns)
        print(df)

    # plot
    if state['format'] in ['line', 'line %']:
        return raw_figures.plot_line(df, **style_kwargs)
    elif state['format'] in ['area', 'area %']:
        return raw_figures.plot_stacked_bar(df, **style_kwargs)
    else:
        raise Exception('invalid format: ' + str(state['format']))



from __future__ import annotations

import os
import typing

from . import paths

if typing.TYPE_CHECKING:
    from mince import types
    import polars as pl


def load_dashboard_data(
    *,
    data_dir: str | None = None,
    dashboard: str | None = None,
    job_start_time: int | None = None,
) -> tuple[types.CollectionJobSummary, dict[str, pl.DataFrame]]:
    import polars as pl

    # load job
    if job_start_time is None:
        job = read_latest_job_summary(dashboard=dashboard, data_dir=data_dir)
        if job is None:
            raise Exception('no collection jobs have been run')
    else:
        job = read_job_summary(
            dashboard=dashboard, data_dir=data_dir, timestamp=job_start_time
        )

    # load data
    if data_dir is None:
        data_dir = paths.get_dashboard_data_dir(dashboard=job['dashboard_name'])
    dfs = {}
    for datum_name in job['data_names']:
        path = paths.get_job_datum_path(
            datum_name=datum_name,
            data_dir=data_dir,
            timestamp=job['job_start_time'],
        )
        dfs[datum_name] = pl.read_parquet(path)

    return job, dfs


def read_job_summary(
    *,
    dashboard: str | None = None,
    data_dir: str | None = None,
    timestamp: int | None = None,
) -> types.CollectionJobSummary:
    raise Exception()


def read_latest_job_summary(
    *,
    dashboard: str | None = None,
    data_dir: str | None = None,
) -> types.CollectionJobSummary | None:
    import json

    # get path
    latest_path = paths.get_job_latest_path(
        dashboard=dashboard, data_dir=data_dir
    )

    # load data
    if not os.path.isfile(latest_path):
        return None
    else:
        with open(latest_path, 'r') as f:
            job: types.CollectionJobSummary = json.load(f)
            return job



from __future__ import annotations

from typing import Any, TYPE_CHECKING

from mince import dashboards
from mince import specs

if TYPE_CHECKING:
    import polars as pl
    from mince import types


def _get_timeseries_style_kwargs(
    state: dict[str, Any],
    df: pl.DataFrame,
    data_date_range: tuple[int, int],
    ui_spec: types.UiSpec,
    title: str | None = None,
) -> dict[str, Any]:
    # get yscale
    log_y = False
    if state['format'] == 'line' and state['yscale'] == 'log':
        log_y = True

    relative_y = False
    if state['format'] == 'line' and state['yscale'] == 'relative':
        relative_y = True

    # get ytick format
    df_metric = dashboards.get_df_metric(state=state, ui_spec=ui_spec)
    metric_spec = ui_spec['metrics'][df_metric]
    if state['format'] in ['line %', 'area %']:
        ytick_format = '%'
    elif metric_spec['unit'] in ['usd', 'USD', '$']:
        ytick_format = '$'
    elif metric_spec['unit'] == 'bytes':
        ytick_format = 'bytes'
    else:
        ytick_format = None

    # get title
    if title is None:
        title = _create_timeseries_title(
            df=df, state=state, data_date_range=data_date_range, ui_spec=ui_spec
        )

    # get ylabel
    ylabel = title.lower()
    if state['format'] in ['line %', 'area %']:
        ylabel = ylabel + ' share'

    # get xlabel
    if state['sample_interval'] == 'month':
        xlabel = 'month'
    elif state['sample_interval'] == 'week':
        xlabel = 'week'
    elif state['sample_interval'] == 'date':
        xlabel = 'date'
    else:
        raise Exception(
            'invalid selected sample_interval: ' + str(state['sample_interval'])
        )

    x = 'timestamp'
    total = False
    hover = True

    colors = ui_spec.get('colors')

    if state['format'] == 'line':
        # toggle total entry
        if state['total'] == 'total':
            total = True
        elif state['total'] == 'no total':
            total = False
        else:
            raise Exception('unknown value for total: ' + str(state['total']))

        # toggle hover
        if state['hover'] == 'hover':
            hover = True
        elif state['hover'] == 'no hover':
            hover = False
        else:
            raise Exception('unknown value for hover: ' + str(state['hover']))

        # age-specific styles
        if state['xalign'] == 'age':
            if state['sample_interval'] == 'date':
                xlabel = 'days'
            else:
                xlabel = xlabel + 's'
            xlabel = state['grouping'] + ' age (' + xlabel + ')'
            x = 'age'

        # relative styles
        if state['ynormalize'] == 'relative':
            ytick_format = '%'

    # return options based on figure type
    df_metric = dashboards.get_df_metric(state, ui_spec)
    if state['format'] in ['line', 'line %']:
        return dict(
            metric=df_metric,
            grouping=state['grouping'],
            x=x,
            log_y=log_y,
            relative_y=relative_y,
            xlabel=xlabel,
            ylabel=ylabel,
            ytick_format=ytick_format,
            title=title,
            hover=hover,
            total=total,
            colors=colors,
        )
    elif state['format'] in ['area', 'area %']:
        styles: dict[str, Any] = dict(
            # metric=state['metric'],
            metric=df_metric,
            grouping=state['grouping'],
            x=x,
            xlabel=xlabel,
            ylabel=ylabel,
            ytick_format=ytick_format,
            title=title,
            colors=colors,
            interval=state['sample_interval'],
        )
        if state['format'] == 'area %':
            styles['ylim'] = [0, 1]
        return styles
    else:
        raise Exception('invalid format')


def _create_timeseries_title(
    df: pl.DataFrame,
    state: dict[str, Any],
    data_date_range: tuple[int, int],
    ui_spec: types.UiSpec,
) -> str:
    df_metric = dashboards.get_df_metric(state, ui_spec)
    if df_metric in ui_spec['submetrics']:
        submetric = specs.state_to_alias(
            state['submetric'], 'submetric', ui_spec
        ).lower()
        title: str = submetric
    else:
        title = specs.state_to_alias(df_metric, 'metric', ui_spec).lower()

    title_prefix = ui_spec['title_prefix']
    title_infix = ui_spec['title_infix']
    title_postfix = ui_spec['title_postfix']

    if title_infix is not None:
        title = title_infix + ' ' + title
    if title_postfix is not None:
        title = title + ' ' + title_postfix

    if (
        state['cumulative'] in ['cumulative (window)', 'cumulative (t=0)']
        and ui_spec['metrics'][df_metric]['type'] == 'sum'
    ):
        title = 'Cumulative ' + title
    else:
        title = (
            specs.state_to_alias(
                state['sample_interval'], 'sample_interval', ui_spec
            ).title()
            + ' '
            + title
        )

    if title_prefix is not None:
        title = title_prefix + title

    if (
        state['cumulative'] in ['cumulative (window)', 'cumulative (t=0)']
        and ui_spec['metrics'][df_metric]['type'] == 'sum'
    ):
        import tooltime

        if state['cumulative'] == 'cumulative (t=0)':
            start_date = tooltime.timestamp_to_date(data_date_range[0])
        elif state['cumulative'] == 'cumulative (window)':
            start_date = tooltime.timestamp_to_date(df['timestamp'].min())  # type: ignore
        else:
            raise Exception()
        end_date = tooltime.timestamp_to_date(df['timestamp'].max())  # type: ignore
        title += ' from ' + start_date + ' to ' + end_date

    return title.title()



from __future__ import annotations

import argparse
import typing

import mince
from ... import ops
from .. import cli_helpers
from .. import cli_parsing

if typing.TYPE_CHECKING:
    from mince import types


def add_collect_command(
    subparsers: cli_helpers.Subparsers,
    include_dashboard_arg: bool,
    DashboardClass: typing.Type[types.Dashboard] | None,
) -> None:
    parser = subparsers.add_parser(
        'collect',
        help='collect data of dashboard',
        formatter_class=cli_helpers.HelpFormatter,
        allow_abbrev=False,
    )
    if include_dashboard_arg:
        parser.add_argument('dashboard', help='dashboard to collect data of')
    parser.add_argument('--data-dir', help='data directory')
    parser.add_argument(
        '--start', help='start date to use [default = ETH_GENESIS]'
    )
    parser.add_argument(
        '--end',
        help='end date to use [default = now]',
    )
    parser.add_argument(
        '--intervals',
        nargs='+',
        metavar='INTERVAL',
        help='intervals to collect: {day week month}',
    )
    parser.add_argument(
        '--skip-incomplete-intervals',
        help='skip incomplete time intervals instead of collecting them',
        action='store_true',
    )
    parser.add_argument(
        '--extra-kwargs', metavar='JSON', help='raw extra kwargs, as json'
    )
    parser.add_argument(
        '--extra-kwargs-update',
        metavar='JSON',
        help='raw extra kwargs update, as json',
    )
    parser.add_argument('--dry', help='dry run', action='store_true')
    parser.add_argument(
        '--verbose',
        help='verbosity level',
        nargs='?',
        default=None,
        type=int,
        const=1,
    )

    if DashboardClass is not None:
        custom_args = get_custom_cli_args(DashboardClass)
        for name, spec in custom_args.items():
            parser.add_argument(*name, **spec)

    parser.set_defaults(f_command=collect_command)
    cli_parsing.add_common_args(parser)


async def collect_command(
    args: argparse.Namespace,
    dashboard: str | typing.Type[types.Dashboard] | None = None,
) -> None:
    # parse registry
    registry: types.RegistryFile = {
        'path': args.registry,
        'validate': not args.no_validate,
    }

    # parse dashboard class
    if dashboard is None:
        dashboard = args.dashboard
    if isinstance(dashboard, str):
        DashboardClass = mince.dashboards.get_dashboard_class(
            name=dashboard, registry=registry
        )
    elif type(dashboard) is type and issubclass(dashboard, mince.Dashboard):
        DashboardClass = dashboard
    else:
        raise Exception()

    # collect input CollectKwargs
    input_collect_kwargs = get_input_collect_kwargs(args, DashboardClass)

    # parse custom extra_kwargs arguments
    collect_kwargs = ops.collection.collection_run._resolve_collect_kwargs(
        dashboard_name=DashboardClass.load_spec()['name'],
        dashboard_class=DashboardClass,
        class_collect_kwargs=DashboardClass.get_default_collect_kwargs(),
        direct_collect_kwargs=input_collect_kwargs,
        registry_entry=None,  # TODO: include registry entry
    )

    await ops.async_collect_dashboard_data(
        dashboard=dashboard,
        registry=registry,
        **collect_kwargs,
    )


def get_custom_cli_args(
    DashboardClass: type[mince.Dashboard],
) -> dict[tuple[str], dict[str, typing.Any]]:
    collect_kwargs = ops.collection.collection_run._resolve_collect_kwargs(
        dashboard_name=DashboardClass.load_spec()['name'],
        dashboard_class=DashboardClass,
        class_collect_kwargs=DashboardClass.get_default_collect_kwargs(),
        direct_collect_kwargs={},
        registry_entry=None,  # TODO: include registry entry
    )
    args = {}
    for key, value in collect_kwargs['extra_kwargs'].items():
        name = '--' + key.replace('_', '-')
        help_str = key.replace('_', ' ') + ' collection parameter'
        metavar = 'VALUE'
        if isinstance(value, bool):
            if value:
                name = name.replace('--', '--no-')
            spec = dict(
                help=help_str,
                action='store_const',
                const=True,
                default=None,
            )
        elif isinstance(value, (str, int, float)):
            spec = dict(help=help_str, metavar=metavar, type=type(value))
        elif isinstance(value, list):
            spec = dict(help=help_str, metavar=metavar, nargs='*')
        else:
            raise Exception('invalid extra_args type for cli: ' + str())
        args[(name,)] = spec
    return args


def get_input_collect_kwargs(
    args: argparse.Namespace,
    DashboardClass: type[mince.Dashboard],
) -> mince.CollectKwargsPartial:
    import json

    # parse extra_kwargs
    if args.extra_kwargs is not None:
        extra_kwargs = json.loads(args.extra_kwargs)
    else:
        extra_kwargs = None
    if args.extra_kwargs_update is not None:
        extra_kwargs_update = json.loads(args.extra_kwargs_update)
    else:
        extra_kwargs_update = {}

    custom_cli_args = get_custom_cli_args(DashboardClass)
    for names, spec in custom_cli_args.items():
        input_name = names[-1].lstrip('-').replace('-', '_')
        if input_name.startswith('no_'):
            extra_kwargs_update[input_name[3:]] = not getattr(args, input_name)
        else:
            extra_kwargs_update[input_name] = getattr(args, input_name)

    return {
        'data_dir': args.data_dir,
        'start_time': args.start,
        'end_time': args.end,
        'intervals': args.intervals,
        'skip_incomplete_intervals': args.skip_incomplete_intervals,
        'extra_kwargs': extra_kwargs,
        'extra_kwargs_update': extra_kwargs_update,
        'dry': args.dry,
        'verbose': args.verbose,
    }



from __future__ import annotations

import typing

from mince import storage
from .dashboard_class import Dashboard

if typing.TYPE_CHECKING:
    from typing import Type
    from mince import types


def get_dashboard_class(
    name: str, *, registry: types.RegistryReference = None
) -> Type[Dashboard]:
    """syntax:

    module.submodule.DashboardClass
    OR
    registered_name
    """
    if '.' in name:
        return resolve_dashboard_class(name)
    else:
        entry = storage.get_registry_entry(name=name, registry=registry)
        return resolve_dashboard_class(entry['dashboard_class'])


def resolve_dashboard_class(reference: str) -> Type[Dashboard]:
    """syntax:

    module.submodule.DashboardClassName
    OR
    module (will try module.mince)
    """
    import importlib

    *module_name_pieces, class_name = reference.split('.')
    module_name = '.'.join(module_name_pieces)
    module = importlib.import_module(module_name)
    DashboardClass: type[Dashboard] = getattr(module, class_name)
    if not issubclass(DashboardClass, Dashboard):
        raise Exception('not a subclass of Dashboard')
    return DashboardClass


def find_dashboard_class(package: str) -> Type[Dashboard]:
    import importlib

    module_name = package + '.mince'
    module = importlib.import_module(module_name)
    candidates = []
    for name, value in vars(module).items():
        if type(value) is type and issubclass(value, Dashboard):
            candidates.append(value)
    if len(candidates) == 0:
        raise Exception('no Dashboard classes in ' + module_name)
    elif len(candidates) > 1:
        raise Exception('multiple Dashboard classes in ' + module_name)
    else:
        return candidates[0]


def instantiate_dashboard(
    dashboard: str | Type[Dashboard],
    *,
    registry: types.RegistryReference = None,
    instance_kwargs: typing.Mapping[str, typing.Any] | None = None,
) -> Dashboard:
    # load data
    data_dir = storage.resolve_data_dir(dashboard, registry=registry)
    job, dfs = storage.load_dashboard_data(data_dir=data_dir)

    # get dashboard class
    if isinstance(dashboard, str):
        DashboardClass = get_dashboard_class(dashboard, registry=registry)
    elif issubclass(dashboard, Dashboard):
        DashboardClass = dashboard
    else:
        raise Exception()

    # load spec
    ui_spec = DashboardClass.load_spec()

    # create dashboard instance
    date_range = (
        job['collect_kwargs']['start_time'],
        job['collect_kwargs']['end_time'],
    )
    if instance_kwargs is None:
        instance_kwargs = {}
    return DashboardClass(
        dfs=dfs, spec=ui_spec, job=job, date_range=date_range, **instance_kwargs
    )



from __future__ import annotations

import argparse
import typing

import mince
from ... import ops
from .. import cli_helpers
from .. import cli_parsing

if typing.TYPE_CHECKING:
    from mince import types


def add_run_command(
    subparsers: cli_helpers.Subparsers,
    include_dashboard_arg: bool,
    DashboardClass: typing.Type[mince.Dashboard] | None,
) -> None:
    parser = subparsers.add_parser(
        'run',
        help='run dashboard',
        formatter_class=cli_helpers.HelpFormatter,
    )
    parser.set_defaults(f_command=run_command)
    if include_dashboard_arg:
        parser.add_argument(
            'dashboard',
            help='dashboard to run',
        )
    parser.add_argument(
        '--port',
        help='port number to run dashboard on',
    )
    cli_parsing.add_common_args(parser)


def run_command(
    args: argparse.Namespace,
    dashboard: str | typing.Type[types.Dashboard] | None = None,
) -> None:
    registry: types.RegistryFile = {
        'path': args.registry,
        'validate': not args.no_validate,
    }
    if dashboard is None:
        dashboard = args.dashboard
    ops.run_dashboard(
        dashboard=dashboard,
        registry=registry,
        run_kwargs={'port': args.port},
        instance_kwargs={'pdb': args.pdb},
    )



from __future__ import annotations

import argparse
import typing

import mince
from ... import dashboards
from ... import storage
from .. import cli_helpers
from .. import cli_parsing

if typing.TYPE_CHECKING:
    from mince import types


def add_validate_command(
    subparsers: cli_helpers.Subparsers,
    include_dashboard_arg: bool,
    DashboardClass: typing.Type[mince.Dashboard] | None,
) -> None:
    parser = subparsers.add_parser(
        'validate',
        help='validate dashboard spec and data',
        formatter_class=cli_helpers.HelpFormatter,
    )
    parser.set_defaults(f_command=validate_command)
    if include_dashboard_arg:
        parser.add_argument(
            'dashboard',
            help='dashboard to validate',
            nargs='?',
        )
    parser.add_argument(
        '--spec',
        help='validate spec specifically',
        action='store_true',
    )
    parser.add_argument(
        '--data',
        help='validate data specifically',
        action='store_true',
    )
    parser.add_argument(
        '--registry',
        help='validate registry specifically',
        action='store_true',
    )
    cli_parsing.add_common_args(parser, registry_name='--registry-path')


def validate_command(
    args: argparse.Namespace,
    dashboard: str | typing.Type[types.Dashboard] | None = None,
) -> None:
    # determine validation targets
    if not args.registry and not args.spec and not args.data:
        targets = ['registry', 'spec', 'data']
    else:
        targets = []
        if args.registry:
            targets.append('registry')
        if args.spec:
            targets.append('spec')
        if args.data:
            targets.append('data')

    # validate registry
    if 'registry' in targets:
        storage.resolve_registry(args.registry_path)

    # validate each dashboard
    if args.dashboard is None and dashboard is None:
        # validate all dashboards
        registry = storage.resolve_registry(args.registry_path)
        names = list(registry['dashboards'].keys())
        if len(registry['dashboards']) == 0:
            print('[no registered dashboards]')
        elif len(registry['dashboards']) > 2:
            print('validating', len(names), 'dashboards:', ', '.join(names))
        for name, entry in registry['dashboards'].items():
            validate_dashboard(
                name=entry['name'],
                dashboard=entry['dashboard_class'],
                targets=targets,
                registry=args.registry_path,
            )
    elif dashboard is not None:
        if type(dashboard) is type and issubclass(dashboard, types.Dashboard):
            validate_dashboard(
                name=dashboard.__name__,
                dashboard=dashboard,
                targets=targets,
                registry=args.registry_path,
            )
        elif isinstance(dashboard, str):
            validate_dashboard(
                name=dashboard,
                targets=targets,
                registry=args.registry_path,
            )
        else:
            raise Exception(
                'invalid type for dashboard: ' + str(type(dashboard))
            )
    else:
        validate_dashboard(
            name=args.dashboard,
            targets=targets,
            registry=args.registry_path,
        )


def validate_dashboard(
    name: str,
    *,
    dashboard: str | typing.Type[types.Dashboard] | None = None,
    targets: list[str],
    registry: str | None,
) -> None:
    import mince.schemas
    import mince.types

    print('validating', dashboard)

    if dashboard is not None:
        if isinstance(dashboard, str):
            DashboardClass = dashboards.get_dashboard_class(
                dashboard, registry=registry
            )
        elif type(dashboard) is type and issubclass(dashboard, types.Dashboard):
            DashboardClass = dashboard
        else:
            raise Exception('invalid type for dashboard')
    else:
        DashboardClass = dashboards.get_dashboard_class(name, registry=registry)

    if 'spec' in targets or 'data' in targets:
        spec = DashboardClass.load_spec()
    if 'spec' in targets:
        mince.types.validate_typeddict(spec, mince.types.UiSpec)
        print('    validated', name, 'spec')
    if 'data' in targets:
        if dashboard is not None:
            data_dir = storage.resolve_data_dir(dashboard)
        else:
            data_dir = storage.resolve_data_dir(name)
        job, dfs = storage.load_dashboard_data(data_dir=data_dir)
        mince.schemas.validate_data_matches_spec(dfs=dfs, spec=spec)
        print('    validated', name, 'data')



"""
Collection parameters come from 3 places (in this order of precedence):
1. arguments passed directly to collect_dashboard_data()
2. arguments stored in the dashboard registry
3. arguments returned by DashboardClass.get_default_collect_kwargs()
"""

from __future__ import annotations

import typing

from mince import dashboards
from mince import schemas
from mince import specs
from mince import storage

if typing.TYPE_CHECKING:
    from typing import Type

    T = typing.TypeVar('T')
    import polars as pl
    import tooltime
    from mince import types


async def async_collect_dashboard_data(
    dashboard: str | Type[dashboards.Dashboard],
    *,
    registry: types.RegistryReference = None,
    data_dir: str | None = None,
    start_time: tooltime.Timestamp | None = None,
    end_time: tooltime.Timestamp | None = None,
    intervals: list[types.Interval] | None = None,
    skip_incomplete_intervals: bool | None = None,
    extra_kwargs: dict[str, typing.Any] | None = None,
    extra_kwargs_update: dict[str, typing.Any] | None = None,
    dry: bool | None = None,
    verbose: int | None = None,
) -> dict[str, pl.DataFrame]:
    import time
    import tooltime

    # gather metadata
    # (must gather before collection because want pre-start git commit versions)
    job_start_time = time.time()
    mince_version = specs.get_package_version('mince')

    # get dashboard class
    (
        dashboard_class,
        dashboard_name,
        dashboard_version,
        registry_entry,
    ) = _resolve_dashboard(dashboard=dashboard, registry=registry)

    # gather collect kwargs
    collect_kwargs = _resolve_collect_kwargs(
        dashboard_name=dashboard_name,
        dashboard_class=dashboard_class,
        class_collect_kwargs=dashboard_class.get_default_collect_kwargs(),
        registry_entry=registry_entry,
        direct_collect_kwargs=dict(
            data_dir=data_dir,
            start_time=start_time,
            end_time=end_time,
            intervals=intervals,
            skip_incomplete_intervals=skip_incomplete_intervals,
            extra_kwargs=extra_kwargs,
            extra_kwargs_update=extra_kwargs_update,
            dry=dry,
            verbose=verbose,
        ),
    )

    # print summary
    if collect_kwargs['verbose'] >= 1:
        _print_collection_summary(
            dashboard_class=dashboard_class,
            dashboard_name=dashboard_name,
            dashboard_version=dashboard_version,
            mince_version=mince_version,
            collect_kwargs=collect_kwargs,
            job_start_time=job_start_time,
        )

    # collect data
    dfs = await dashboard_class.async_collect_data(**collect_kwargs)
    job_end_time = time.time()

    # validate data
    spec = dashboard_class.load_spec()
    schemas.validate_data_matches_spec(dfs, spec)

    # create job metadata
    job: types.CollectionJobSummary = {
        'mince_version': mince_version,
        'dashboard_version': dashboard_version,
        'dashboard_name': dashboard_name,
        'dashboard_class': (
            dashboard_class.__module__ + '.' + dashboard_class.__name__
        ),
        'job_start_time': job_start_time,
        'job_end_time': job_end_time,
        'data_names': list(dfs.keys()),
        'collect_kwargs': collect_kwargs,
    }

    # write results
    if not dry:
        storage.write_job_results(job=job, dfs=dfs)

    # print summary
    if collect_kwargs['verbose'] >= 1:
        print()
        duration = '%.02f' % (job_end_time - job_start_time)
        print(
            'collection completed in',
            duration,
            'seconds at',
            tooltime.timestamp_to_iso_pretty(job_end_time),
        )

    return dfs


def _print_collection_summary(
    *,
    dashboard_class: Type[types.Dashboard],
    dashboard_name: str,
    dashboard_version: str,
    mince_version: str,
    collect_kwargs: types.CollectKwargs,
    job_start_time: tooltime.Timestamp,
) -> None:
    import toolstr
    import tooltime

    toolstr.print_text_box('collecting data')
    print()
    toolstr.print_header('Dashboard')
    print('- name:', dashboard_name)
    print(
        '- class:', dashboard_class.__module__ + '.' + dashboard_class.__name__
    )
    print('- dashboard version:', dashboard_version)
    print('- mince version:', mince_version)
    print()
    toolstr.print_header('Collection')
    for key, value in collect_kwargs.items():
        if key in ['start_time', 'end_time']:
            print('-', key + ':', tooltime.timestamp_to_iso_pretty(value))  # type: ignore
        elif isinstance(value, dict):
            print('-', key + ':')
            for subkey, subvalue in value.items():
                print('    -', subkey + ':', subvalue)
        else:
            print('-', key + ':', value)
    print('- job start time:', tooltime.timestamp_to_iso_pretty(job_start_time))


def _validate_data(
    dfs: typing.Any,
    DashboardClass: typing.Type[types.Dashboard],
) -> None:
    import polars as pl

    # check type
    if (not isinstance(dfs, dict)) or any(
        not isinstance(name, str) or not isinstance(df, pl.DataFrame)
        for name, df in dfs.items()
    ):
        raise Exception('invalid format for data')


def _resolve_dashboard(
    dashboard: str | Type[dashboards.Dashboard],
    registry: types.RegistryReference = None,
) -> tuple[Type[types.Dashboard], str, str, types.RegistryEntry | None]:
    if isinstance(dashboard, str):
        dashboard_class = dashboards.get_dashboard_class(
            dashboard, registry=registry
        )
        dashboard_name = dashboard
        entry = storage.get_registry_entry(dashboard, registry=registry)
    elif type(dashboard) is type and issubclass(
        dashboard, dashboards.Dashboard
    ):
        dashboard_class = dashboard
        dashboard_name = None
        entry = None
    else:
        raise Exception('invalid format for dashboard')

    spec = dashboard_class.load_spec()
    if dashboard_name is None:
        dashboard_name = spec['name']
    version = spec['version']

    return dashboard_class, dashboard_name, version, entry


def _resolve_collect_kwargs(
    dashboard_name: str,
    dashboard_class: Type[dashboards.Dashboard],
    class_collect_kwargs: types.CollectKwargsPartial,
    registry_entry: types.RegistryEntry | None,
    direct_collect_kwargs: types.CollectKwargsPartial,
) -> types.CollectKwargs:
    import tooltime

    # gather all kwargs sets
    if registry_entry is not None:
        registry_collect_kwargs = registry_entry['collect_kwargs']
    else:
        registry_collect_kwargs = {}
    all_kwargs = [
        direct_collect_kwargs,
        registry_collect_kwargs,
        class_collect_kwargs,
    ]

    # data_dir
    data_dir = first_non_none([kwargs.get('data_dir') for kwargs in all_kwargs])
    if data_dir is None:
        data_dir = storage.get_dashboard_data_dir(dashboard_name)

    # start_time
    raw_start_time = first_non_none(  # type: ignore
        [kwargs.get('start_time') for kwargs in all_kwargs]
    )
    if raw_start_time is None:
        raise Exception('raw_start_time not specified')
    start_time = tooltime.timestamp_to_seconds_precise(raw_start_time)

    # end_time
    raw_end_time = first_non_none(  # type: ignore
        [kwargs.get('end_time') for kwargs in all_kwargs]
    )
    if raw_end_time is None:
        raise Exception('raw_end_time not specified')
    end_time = tooltime.timestamp_to_seconds_precise(raw_end_time)

    # intervals
    intervals = first_non_none(
        [kwargs.get('intervals') for kwargs in all_kwargs]
    )
    if intervals is None:
        raise Exception('intervals not specified')

    # skip_incomplete_intervals
    skip_incomplete_intervals = first_non_none(
        [kwargs.get('skip_incomplete_intervals') for kwargs in all_kwargs]
    )
    if skip_incomplete_intervals is None:
        raise Exception('skip_incomplete_intervals not specified')

    # extra_kwargs
    extra_kwargs = first_non_none(
        [kwargs.get('extra_kwargs') for kwargs in all_kwargs]
    )
    if extra_kwargs is None:
        raise Exception('extra_kwargs not specified')
    extra_kwargs = extra_kwargs.copy()

    # extra kwargs update
    for kwargs in all_kwargs:
        extra_kwargs_update = kwargs.get('extra_kwargs_update')
        if extra_kwargs_update is not None:
            for key, value in extra_kwargs_update.items():
                if extra_kwargs.get(key) is None:
                    extra_kwargs[key] = value

    # dry
    dry = first_non_none([kwargs.get('dry') for kwargs in all_kwargs])
    if dry is None:
        raise Exception('dry not specified')

    # verbose
    verbose = first_non_none([kwargs.get('verbose') for kwargs in all_kwargs])
    if verbose is None:
        raise Exception('verbose not specified')

    return {
        'data_dir': data_dir,
        'start_time': start_time,
        'end_time': end_time,
        'intervals': intervals,
        'skip_incomplete_intervals': skip_incomplete_intervals,
        'extra_kwargs': extra_kwargs,
        'dry': dry,
        'verbose': verbose,
    }


def first_non_none(lst: list[T | None]) -> T | None:
    return next((x for x in lst if x is not None), None)



from __future__ import annotations

import mince

import typing

if typing.TYPE_CHECKING:
    import tooltime
    import polars as pl


def create_piece_path(
    data_dir: str,
    path_template: str,
    interval: str,
    tags: dict[str, str],
    timestamp: tooltime.Timestamp | None,
    complete: bool | str,
) -> str:
    import os
    import tooltime

    if isinstance(complete, bool):
        if complete:
            complete_str = 'complete'
        else:
            complete_str = 'incomplete'
    elif isinstance(complete, str):
        if complete not in ['complete', 'incomplete', '*']:
            raise Exception('invalid value for complete parameter')
        complete_str = complete

    if timestamp is not None and timestamp != '*':
        dt = tooltime.timestamp_to_datetime(timestamp)
        if interval in ['hour', 'minute', 'second']:
            timestamp = dt.strftime('%Y-%m-%d_%H-%M-%S')
        elif interval in ['day', 'week']:
            timestamp = dt.strftime('%Y-%m-%d')
        elif interval == 'month':
            timestamp = dt.strftime('%Y-%m')
        elif interval == 'year':
            timestamp = dt.strftime('%Y')
        else:
            raise Exception('invalid interval')
    path = path_template.format(
        interval=interval, timestamp=timestamp, complete=complete_str, **tags
    )
    return os.path.join(data_dir, 'pieces', path)


def pieces_already_exist(
    path_template: str,
    data_dir: str,
    interval: mince.Interval,
    tags: dict[str, str],
    skip_incomplete_intervals: bool,
    start_time: float,
    end_time: float,
    verbose: int,
) -> bool:
    import os
    import tooltime

    duration = {'day': '1d', 'week': '1w', 'month': '1M', 'year': '1y'}.get(
        interval, interval
    )
    chunks = tooltime.get_intervals(
        interval=duration,
        start=start_time,
        end=end_time,
        include_incomplete=not skip_incomplete_intervals,
    )

    all_exist = True
    for timestamp in chunks['start']:
        path = create_piece_path(
            path_template=path_template,
            data_dir=data_dir,
            interval=interval,
            tags=tags,
            timestamp=timestamp,
            complete=True,
        )
        exists = os.path.exists(path)
        if not exists:
            if verbose <= 2:
                return False
            else:
                all_exist = False

        # print summary
        if verbose >= 3:
            if exists:
                label = 'EXISTS'
            else:
                label = 'MISSING'
            print(label, os.path.relpath(path, data_dir))

    return all_exist


def write_piece_files(
    path_template: str,
    data_dir: str,
    df: pl.DataFrame,
    interval: mince.Interval,
    tags: dict[str, str],
) -> None:
    import datetime
    import os
    import shutil

    partitions = df.partition_by(('timestamp', 'complete'), as_dict=True)
    for (timestamp, complete), sub_df in partitions.items():
        path = create_piece_path(
            path_template=path_template,
            data_dir=data_dir,
            interval=interval,
            tags=tags,
            timestamp=typing.cast(datetime.datetime, timestamp),
            complete=typing.cast(bool, complete),
        )
        os.makedirs(os.path.dirname(path), exist_ok=True)
        path_tmp = path + '_tmp'
        sub_df.write_parquet(path_tmp)
        shutil.move(path_tmp, path)

        # delete incomplete version
        if complete:
            incomplete_path = create_piece_path(
                path_template=path_template,
                data_dir=data_dir,
                interval=interval,
                tags=tags,
                timestamp=typing.cast(datetime.datetime, timestamp),
                complete=False,
            )
            if os.path.exists(incomplete_path):
                os.remove(incomplete_path)


def read_piece_files(
    data_dir: str,
    path_template: str,
    tag_names: typing.Sequence[str],
) -> dict[str, pl.DataFrame]:
    import polars as pl

    path = create_piece_path(
        data_dir=data_dir,
        path_template=path_template,
        interval='*',
        tags={tag_name: '*' for tag_name in tag_names},
        timestamp='*',
        complete='*',
    )

    return {'data': pl.read_parquet(path)}



from __future__ import annotations

from typing import Mapping, Sequence, TYPE_CHECKING

import mince.schemas
from . import spec_create

if TYPE_CHECKING:
    import polars as pl
    from mince import types


def create_default_ui_spec(
    *,
    name: str,
    title: str | None = None,
    description: str | None = None,
    favicon_emoji: str | None = None,
    version: str,
    df: pl.DataFrame | None = None,
    metrics: dict[str, types.MetricSpec],
    default_inputs: dict[str, mince.StateDatum | None],
    metric_aliases: dict[str, types.StateDatum],
    submetrics: Mapping[str, Sequence[str]],
    grouping_aliases: dict[str, str],
    colors: dict[str, str],
    add_shortcuts: dict[str, types.ShortcutSpec] | None = None,
    schema: types.DataSchemaPartial | None = None,
    title_prefix: str | None = None,
    title_infix: str | None = None,
    title_postfix: str | None = None,
) -> types.UiSpec:
    inputs, metric_selectors = _get_default_inputs(
        metric_aliases=metric_aliases,
        metrics=metrics,
        grouping_aliases=grouping_aliases,
        submetrics=submetrics,
        default_inputs=default_inputs,
    )

    if metric_aliases is not None:
        inputs['metric']['aliases'] = metric_aliases
        inputs['submetric']['aliases'] = metric_aliases

    shortcuts = _get_default_shortcuts()
    if add_shortcuts is not None:
        shortcuts.update(add_shortcuts)

    if schema is None:
        schema = {'columns': mince.schemas.get_default_columns()}  # type: ignore
    schema_whole = mince.schemas.partial_schema_to_whole(schema)

    return spec_create.create_ui_spec(
        name=name,
        title=title,
        description=description,
        favicon_emoji=favicon_emoji,
        version=version,
        groupings=list(grouping_aliases.values()),
        invalid_states=_get_default_invalid_states(
            metrics=metrics,
            metric_selectors=metric_selectors,
        ),
        inputs=inputs,
        metrics=metrics,
        submetrics=submetrics,
        shortcuts=shortcuts,
        colors=colors,
        schema=schema_whole,
        title_prefix=title_prefix,
        title_infix=title_infix,
        title_postfix=title_postfix,
    )


def _get_default_invalid_states(
    metrics: dict[str, types.MetricSpec],
    metric_selectors: dict[str, dict[str, types.StateDatum]],
) -> list[dict[str, types.StateDatum]]:
    combos: list[dict[str, types.StateDatum]] = []

    line_only_values: list[dict[str, types.StateDatum]] = [
        {'yscale': 'log'},
        {'xalign': 'age'},
        {'ynormalize': 'relative'},
        {'total': 'total'},
    ]
    for value in line_only_values:
        for format in ['line %', 'area', 'area %', 'tree']:
            bad = value.copy()
            bad['format'] = format
            combos.append(bad)

    # only use total when metric type == 'sum'
    for metric, metric_spec in metrics.items():
        if metric_spec['type'] != 'sum':
            combo = dict(metric_selectors[metric], total='total')
            combos.append(combo)

    combos.extend(
        [
            {'time_window': '7d', 'sample_interval': 'monthly'},
            {'time_window': '7d', 'sample_interval': 'weekly'},
            {'time_window': '30d', 'sample_interval': 'monthly'},
        ]
    )

    combos.extend(
        [
            {'xalign': 'age', 'ynormalize': 'relative'},
            {'xalign': 'age', 'total': 'total'},
        ]
    )

    return combos


def _get_default_shortcuts() -> dict[str, types.ShortcutSpec]:
    return {
        '[': {'action': 'cycle_previous', 'field': 'metric'},
        ']': {'action': 'cycle_next', 'field': 'metric'},
        '{': {'action': 'cycle_previous', 'field': 'submetric'},
        '}': {'action': 'cycle_next', 'field': 'submetric'},
        'L': {'action': 'select', 'field': 'format', 'value': 'line'},
        'P': {'action': 'select', 'field': 'format', 'value': 'line %'},
        'A': {'action': 'select', 'field': 'format', 'value': 'area'},
        'p': {'action': 'select', 'field': 'format', 'value': 'area %'},
        'T': {'action': 'select', 'field': 'format', 'value': 'tree'},
        'w': {'action': 'select', 'field': 'time_window', 'value': '7d'},
        'm': {'action': 'select', 'field': 'time_window', 'value': '30d'},
        'y': {'action': 'select', 'field': 'time_window', 'value': '365d'},
        'a': {'action': 'select', 'field': 'time_window', 'value': 'all'},
        'D': {'action': 'select', 'field': 'sample_interval', 'value': 'daily'},
        'W': {
            'action': 'select',
            'field': 'sample_interval',
            'value': 'weekly',
        },
        'M': {
            'action': 'select',
            'field': 'sample_interval',
            'value': 'monthly',
        },
        'l': {
            'action': 'cycle_next',
            'field': 'yscale',
            'help': 'toggle log yaxis',
        },
        'r': {
            'action': 'cycle_next',
            'field': 'ynormalize',
            'help': 'toggle relative yxaxis',
        },
        'x': {
            'action': 'cycle_next',
            'field': 'xalign',
            'help': 'toggle age xaxis',
        },
        'ArrowLeft': {
            'action': 'decrement',
            'field': 'now',
            'value': 'large',
            'help': 'move backward in time',
        },
        'ArrowRight': {
            'action': 'increment',
            'field': 'now',
            'value': 'large',
            'help': 'move forward in time',
        },
        'H': {
            'action': 'toggle_ui',
            'field': 'radio_group_visibility',
            'help': 'toggle button visibility',
        },
        '?': {
            'action': 'toggle_ui',
            'field': 'help_open',
            'help': 'toggle help',
        },
        'h': {
            'action': 'toggle_ui',
            'field': 'help_open',
            'help': 'toggle help',
        },
    }


def _get_default_inputs(
    *,
    metric_aliases: dict[str, types.StateDatum],
    metrics: dict[str, types.MetricSpec],
    grouping_aliases: dict[str, str],
    default_inputs: dict[str, mince.StateDatum | None],
    submetrics: Mapping[str, Sequence[str]],
) -> tuple[dict[str, types.InputSpec], dict[str, dict[str, types.StateDatum]]]:
    default_metric = default_inputs['metric']
    if default_metric is None:
        default_metric = list(metrics.keys())[0]
    if not isinstance(default_metric, str):
        raise Exception('could not determine default metric')

    # default time window
    raw_default_time_window = default_inputs.get('time_window')
    if raw_default_time_window is None:
        default_time_window = 'all'
    elif isinstance(raw_default_time_window, str):
        default_time_window = raw_default_time_window
    else:
        raise Exception('invalid type for default_time_window')

    # default format
    raw_default_format = default_inputs.get('format')
    if raw_default_format is None:
        default_format = 'line'
    elif isinstance(raw_default_format, str):
        default_format = raw_default_format
    else:
        raise Exception('invalid type for default_format')

    # yscale
    raw_default_yscale = default_inputs.get('yscale')
    if raw_default_yscale is None:
        default_yscale = 'linear'
    elif isinstance(raw_default_yscale, str):
        default_yscale = raw_default_yscale
    else:
        raise Exception('invalid type for default_yscale')

    # ynormalize
    raw_default_ynormalize = default_inputs.get('ynormalize')
    if raw_default_ynormalize is None:
        default_ynormalize = 'absolute'
    elif isinstance(raw_default_ynormalize, str):
        default_ynormalize = raw_default_ynormalize
    else:
        raise Exception('invalid type for default_ynormalize')

    # xalign
    raw_default_xalign = default_inputs.get('xalign')
    if raw_default_xalign is None:
        default_xalign = 'time'
    elif isinstance(raw_default_xalign, str):
        default_xalign = raw_default_xalign
    else:
        raise Exception('invalid type for default_xalign')

    # cumulative
    raw_default_cumulative = default_inputs.get('cumulative')
    if raw_default_cumulative is None:
        default_cumulative = 'non-cumulative'
    elif isinstance(raw_default_cumulative, str):
        default_cumulative = raw_default_cumulative
    else:
        raise Exception('invalid type for default_cumulative')

    # total
    raw_default_total = default_inputs.get('cumulative')
    if raw_default_total is None:
        default_total = 'no total'
    elif isinstance(raw_default_total, str):
        default_total = raw_default_total
    else:
        raise Exception('invalid type for default_total')

    # hover
    raw_default_hover = default_inputs.get('cumulative')
    if raw_default_hover is None:
        default_hover = 'hover'
    elif isinstance(raw_default_hover, str):
        default_hover = raw_default_hover
    else:
        raise Exception('invalid type for default_hover')

    # default sample interval
    raw_default_sample_interval = default_inputs.get('sample_interval')
    if raw_default_sample_interval is None:
        default_sample_interval = 'weekly'
    elif isinstance(raw_default_sample_interval, str):
        default_sample_interval = raw_default_sample_interval
    else:
        raise Exception('invalid type for default_sample_interval')

    # metric options
    reverse_metric_aliases = {v: k for k, v in metric_aliases.items()}
    metric_options = []
    metric_selectors: dict[str, dict[str, types.StateDatum]] = {}
    for metric in metrics.keys():
        for submetric_category in submetrics.keys():
            if metric in submetrics[submetric_category]:
                metric_selectors[metric] = {
                    'metric': submetric_category,
                    'submetric': reverse_metric_aliases.get(metric, metric),
                }
                if submetric_category not in metric_options:
                    metric_options.append(submetric_category)
                break
        else:
            metric_selectors[metric] = {
                'metric': reverse_metric_aliases.get(metric, metric)
            }
            metric_options.append(reverse_metric_aliases.get(metric, metric))

    # get metric defaults
    for submetric_category in submetrics.keys():
        if default_metric in submetrics[submetric_category]:
            aliased_default_metric = submetric_category
            aliased_submetric_options = [
                reverse_metric_aliases.get(metric, metric)
                for metric in submetrics[submetric_category]
            ]
            aliased_default_submetric = reverse_metric_aliases.get(
                default_metric, default_metric
            )
            break
    else:
        aliased_default_metric = reverse_metric_aliases.get(
            default_metric, default_metric
        )
        if len(submetrics) > 0:
            aliased_submetric_options = list(next(iter(submetrics.values())))
            aliased_default_submetric = aliased_submetric_options[0]
        else:
            aliased_submetric_options = []
            aliased_default_submetric = None

    return {
        'now': {
            'type': 'date',
            'description': 'Select date',
            'default': None,
            'aliases': {},
        },
        'metric': {
            'type': 'button',
            'description': 'Select metric',
            'button_options': metric_options,
            'default': aliased_default_metric,
            'aliases': {},
        },
        'submetric': {
            'type': 'button',
            'description': 'Select submetric',
            'button_options': aliased_submetric_options,
            'default': aliased_default_submetric,
            'visibility': {
                'show_if': [
                    {'metric': metric_aliases.get(metric, metric)}
                    for metric in submetrics.keys()
                ]
            },
            'aliases': {},
        },
        'format': {
            'type': 'button',
            'description': 'Select chart type',
            'button_options': [
                'line',
                'line %',
                'area',
                'area %',
                'tree',
            ],
            'default': default_format,
            'aliases': {},
        },
        'time_window': {
            'type': 'button',
            'description': 'Select time window',
            'button_options': [
                '7d',
                '30d',
                '365d',
                'all',
            ],
            'default': default_time_window,
            'visibility': {
                'hide_if': [
                    dict(metric_selectors[metric], format='tree')
                    for metric, metric_spec in metrics.items()
                    if metric_spec['type'] not in ['sum', 'min', 'max']
                ],
            },
            'aliases': {},
        },
        'sample_interval': {
            'type': 'button',
            'description': 'Select sampling interval',
            'button_options': [
                'daily',
                'weekly',
                'monthly',
            ],
            'default': default_sample_interval,
            'visibility': {
                'hide_if': [
                    dict(metric_selectors[metric], format='tree')
                    for metric, metric_spec in metrics.items()
                    if metric_spec['type'] not in ['unique']
                ],
            },
            'aliases': {
                'daily': 'date',
                'weekly': 'week',
                'monthly': 'month',
            },
        },
        'grouping': {
            'type': 'button',
            'description': 'Select grouping',
            'button_options': list(grouping_aliases.keys()),
            'default': {v: k for k, v in grouping_aliases.items()}[
                default_inputs['grouping']  # type: ignore
            ],
            'aliases': grouping_aliases,  # type: ignore
            'visibility': {
                'hide': len(grouping_aliases) < 2,
            },
        },
        'yscale': {
            'type': 'button',
            'description': 'Select y-axis scale',
            'button_options': [
                'linear',
                'log',
            ],
            'default': default_yscale,
            'visibility': {
                'show_if': [{'format': 'line'}],
                'start_hidden': True,
            },
            'aliases': {},
        },
        'ynormalize': {
            'type': 'button',
            'description': 'Select y-axis normalization',
            'button_options': [
                'absolute',
                'relative',
            ],
            'default': default_ynormalize,
            'visibility': {
                'show_if': [{'format': 'line'}],
                'start_hidden': True,
            },
            'aliases': {},
        },
        'xalign': {
            'type': 'button',
            'description': 'Select x-axis alignment',
            'button_options': [
                'time',
                'age',
            ],
            'default': default_xalign,
            'visibility': {
                'show_if': [{'format': 'line'}],
                'start_hidden': True,
            },
            'aliases': {},
        },
        'cumulative': {
            'type': 'button',
            'description': 'Select whether plot is cumulative',
            'button_options': [
                'non-cumulative',
                'cumulative (window)',
                'cumulative (t=0)',
            ],
            'default': default_cumulative,
            'visibility': {
                'start_hidden': True,
                'show_if': [
                    dict(metric_selectors[metric], format='line')
                    for format in ['line', 'area']
                    for metric, metric_spec in metrics.items()
                    if metric_spec['type'] == 'sum'
                ],
            },
            'aliases': {},
        },
        'total': {
            'type': 'button',
            'description': 'Select whether to show total',
            'button_options': [
                'no total',
                'total',
            ],
            'default': default_total,
            'visibility': {
                'show_if': [
                    dict(metric_selectors[metric], format='line')
                    for metric, metric_spec in metrics.items()
                    if metric_spec['type'] == 'sum'
                ],
                'start_hidden': True,
            },
            'aliases': {},
        },
        'hover': {
            'type': 'button',
            'description': 'Select whether to show mouse hover data',
            'button_options': [
                'hover',
                'no hover',
            ],
            'default': default_hover,
            'visibility': {
                'show_if': [{'format': 'line'}],
                'start_hidden': True,
            },
            'aliases': {},
        },
    }, metric_selectors


def get_package_version(package: str) -> str:
    import importlib
    import os
    import subprocess

    module = importlib.import_module(package)
    try:
        git_branch_cmd = 'git rev-parse --abbrev-ref HEAD'
        git_commit_subcount_cmd = 'git describe --tags --abbrev=0'
        git_commit_count_cmd = 'git rev-list {subcount}..HEAD --count'
        git_tag_cmd = 'git describe --tags --abbrev=0'
        git_status_cmd = 'git diff --numstat'

        cwd = os.path.dirname(module.__file__)  # type: ignore

        tag = (
            subprocess.check_output(
                git_tag_cmd.split(' '), cwd=cwd, stderr=subprocess.DEVNULL
            )
            .decode('utf-8')
            .rstrip('\n')
        )
        subcount = (
            subprocess.check_output(git_commit_subcount_cmd.split(' '), cwd=cwd)
            .decode('utf-8')
            .rstrip('\n')
        )
        commit_count = (
            subprocess.check_output(
                git_commit_count_cmd.format(subcount=subcount).split(' '),
                cwd=cwd,
            )
            .decode('utf-8')
            .rstrip('\n')
        )
        branch = (
            subprocess.check_output(git_branch_cmd.split(' '), cwd=cwd)
            .decode('utf-8')
            .rstrip('\n')
        )
        status = (
            subprocess.check_output(git_status_cmd.split(' '), cwd=cwd)
            .decode('utf-8')
            .strip()
        )

        label = tag
        if commit_count != '0':
            label = tag + ':' + commit_count
        if branch != 'main':
            label = branch + ':' + label
        if len(status) > 0:
            n_adds = 0
            n_removes = 0
            for file in status.split('\n'):
                adds, removes, path = file.split('\t')
                n_adds += int(adds)
                n_removes += int(removes)
            if n_adds > 0:
                label = label + '+' + str(n_adds)
            if n_removes > 0:
                label = label + '-' + str(n_removes)

        return label

    except Exception:
        return module.__version__  # type: ignore



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from mince import types
    import polars as pl


def get_series_columns(schema: dict[str, types.ColumnType]) -> list[str]:
    return [
        column
        for column in schema.keys()
        if column not in ['timestamp', 'value']
    ]


def get_default_columns() -> dict[str, types.ColumnType]:
    import polars as pl

    return {
        'metric': pl.String,
        'timestamp': pl.Datetime(time_unit='ms'),
        'interval': pl.String,  # point, day, week, month
        'value': pl.Float64,
        'complete': pl.Boolean,
    }


def get_blank_dataframe(ui_spec: types.UiSpec) -> pl.DataFrame:
    """get black dataframe conforming to data schema"""
    import polars as pl

    return pl.DataFrame([], schema=ui_spec['schema']['columns'])



from __future__ import annotations

from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    import datetime

    import dash_extensions  # type: ignore
    import dash_bootstrap_components as dbc  # type: ignore
    import tooltime

    from mince import types


time_increments = {
    'all': {
        'large': '365d',
        'medium': '30d',
        'small': '7d',
    },
    '365d': {
        'large': '365d',
        'medium': '30d',
        'small': '7d',
    },
    '30d': {
        'large': '30d',
        'medium': '7d',
        'small': '1d',
    },
    '7d': {
        'large': '7d',
        'medium': '1d',
        'small': '1d',
    },
}


def _create_shortcuts_listeners(
    shortcuts: dict[str, types.ShortcutSpec],
) -> dash_extensions.Keyboard:
    import dash_extensions

    return dash_extensions.Keyboard(
        captureKeys=list(shortcuts.keys()), id='keyboard', n_keydowns=1
    )


def _create_help_modal(
    shortcuts: dict[str, types.ShortcutSpec],
) -> dbc.Modal:
    from dash import html  # type: ignore
    import dash_bootstrap_components as dbc

    rows = []
    for key, shortcut in shortcuts.items():
        field = shortcut['field']
        if field is not None:
            field = field.replace('_', ' ')
        # get description
        if 'help' in shortcut:
            description = shortcut['help']
        elif shortcut['action'] == 'toggle_ui':
            description = 'toggle ' + str(shortcut['field'])
        elif shortcut['action'] == 'select':
            description = html.Div(
                ['set ', field, ' to ', html.B(shortcut['value'])]
            )
        elif shortcut['action'] == 'cycle_next':
            description = html.Div(['go to next ', html.B(field)])
        elif shortcut['action'] == 'cycle_previous':
            description = html.Div(['go to previous ', html.B(field)])
        else:
            raise Exception('invalid shortcut action')

        # get key string
        if key == 'ArrowRight':
            key_str = '→'
        elif key == 'ArrowLeft':
            key_str = '←'
        else:
            key_str = key

        # build table row
        row = [
            html.Td(' '),
            html.Td(html.B(key_str)),
            html.Td(' '),
            html.Td(' '),
            html.Td(' '),
            html.Td(description),
            html.Td(' '),
        ]
        rows.append(row)

    return dbc.Modal(
        [
            dbc.ModalHeader(dbc.ModalTitle('Keyboard Shortcuts')),
            dbc.ModalBody(
                html.Table(html.Tbody([html.Tr(row) for row in rows])),
            ),
        ],
        id='help-modal',
        size='sm',
        is_open=False,
    )


def _process_keyboard_shortcuts(
    display_state: dict[str, Any],
    # help_open: bool,
    ui_state: dict[str, Any],
    raw_shortcut: dict[str, Any],
    data_date_range: tuple[tooltime.Timestamp, tooltime.Timestamp],
    ui_spec: types.UiSpec,
) -> None:
    import tooltime

    shortcut = ui_spec['shortcuts'][raw_shortcut['key']]
    field = shortcut['field']
    if shortcut['action'] == 'select' and field is not None:
        display_state[field] = shortcut['value']
    elif shortcut['action'] == 'cycle_next' and field is not None:
        options = ui_spec['inputs'][field]['button_options']
        index = options.index(display_state[field])
        if index + 1 == len(options):
            display_state[field] = options[0]
        else:
            display_state[field] = options[index + 1]
    elif shortcut['action'] == 'cycle_previous' and field is not None:
        options = ui_spec['inputs'][field]['button_options']
        index = options.index(display_state[field])
        if index == 0:
            display_state[field] = options[-1]
        else:
            display_state[field] = options[index - 1]
    elif (
        shortcut['action'] in ['increment', 'decrement']
        and shortcut['field'] == 'now'
    ):
        if raw_shortcut['ctrlKey'] or raw_shortcut['shiftKey']:
            value = 'medium'
        elif raw_shortcut['ctrlKey'] and raw_shortcut['shiftKey']:
            value = 'small'
        else:
            value = 'large'

        duration = time_increments[display_state['time_window']][value]
        new_now_dt = tooltime.timestamp_to_datetime(display_state['now'][:10])
        if shortcut['action'] == 'increment':
            new_now_dt = _increment_datetime(new_now_dt, duration)
        elif shortcut['action'] == 'decrement':
            new_now_dt = _decrement_datetime(new_now_dt, duration)
        else:
            raise Exception()
        new_now = tooltime.timestamp_to_date(new_now_dt)

        # check that new time is valid
        min_now = tooltime.timestamp_to_date(data_date_range[0])
        if new_now < min_now:
            new_now = min_now
        max_now = tooltime.timestamp_to_date(data_date_range[1])
        if new_now > max_now:
            new_now = max_now

        display_state['now'] = new_now
    elif shortcut['action'] == 'toggle_ui':
        field = shortcut['field']
        if field is None:
            raise Exception('no field for shortcut')
        ui_state[field] = not ui_state[field]
    else:
        raise Exception('invalid shortcut action')


def _increment_datetime(
    dt: datetime.datetime, duration: str
) -> datetime.datetime:
    import datetime

    if duration == '1d':
        return dt + datetime.timedelta(days=1)
    elif duration == '7d':
        return dt + datetime.timedelta(days=7)
    elif duration == '30d':
        if dt.month == 12:
            return dt.replace(month=1, year=dt.year + 1)
        else:
            return dt.replace(month=dt.month + 1)
    elif duration == '365d':
        return dt.replace(year=dt.year + 1)
    else:
        raise Exception('invalid increment')


def _decrement_datetime(
    dt: datetime.datetime, duration: str
) -> datetime.datetime:
    import datetime

    if duration == '1d':
        return dt - datetime.timedelta(days=1)
    elif duration == '7d':
        return dt - datetime.timedelta(days=7)
    elif duration == '30d':
        if dt.month == 1:
            return dt.replace(month=12, year=dt.year - 1)
        else:
            return dt.replace(month=dt.month - 1)
    elif duration == '365d':
        return dt.replace(year=dt.year - 1)
    else:
        raise Exception('invalid increment')



from __future__ import annotations

import argparse
import typing

import mince
from .. import cli_helpers
from .. import cli_parsing


def add_docker_command(
    subparsers: cli_helpers.Subparsers,
    include_dashboard_arg: bool,
    DashboardClass: typing.Type[mince.Dashboard] | None,
) -> None:
    """
    mince docker build <dashboard> [--name name]
    """
    parser = subparsers.add_parser(
        'docker',
        help='performance docker',
        formatter_class=cli_helpers.HelpFormatter,
    )
    parser.set_defaults(f_command=docker_command)
    parser.add_argument('subcommand', help='docker command {build}')
    if include_dashboard_arg:
        parser.add_argument('dashboard', help='dashboard name')
    cli_parsing.add_common_args(parser)


def docker_command(args: argparse.Namespace) -> None:
    command = args.command
    dashboard = args.dashboard
    print('COMMAND', command)
    print('DASHBOARD', dashboard)

    if command == 'build':
        build_dashboard_docker_image()
    else:
        raise Exception()


def build_dashboard_docker_image() -> None:
    pass



from __future__ import annotations

import os
import typing

from . import paths
from . import job_read

if typing.TYPE_CHECKING:
    import polars as pl
    from mince import types


def write_job_results(
    job: types.CollectionJobSummary,
    dfs: dict[str, pl.DataFrame],
    mark_as_latest: bool | None = None,
) -> None:
    import json
    import shutil

    data_dir = job['collect_kwargs']['data_dir']

    # write metadata
    metadata_path = paths.get_job_metadata_path(
        dashboard=job['dashboard_name'],
        timestamp=job['job_start_time'],
        data_dir=data_dir,
    )
    os.makedirs(os.path.dirname(metadata_path), exist_ok=True)
    metadata_path_tmp = metadata_path + '_tmp'
    with open(metadata_path_tmp, 'w') as f:
        json.dump(job, f)
    shutil.move(metadata_path_tmp, metadata_path)

    # write dataframes
    for datum_name, df in dfs.items():
        path = paths.get_job_datum_path(
            datum_name=datum_name,
            data_dir=data_dir,
            timestamp=job['job_start_time'],
        )
        path_tmp = path + '_tmp'
        df.write_parquet(path_tmp)
        shutil.move(path_tmp, path)

    # write latest.json file
    if mark_as_latest is None:
        latest_job = job_read.read_latest_job_summary(
            dashboard=job['dashboard_name']
        )
        if latest_job is None:
            mark_as_latest = True
        else:
            mark_as_latest = (
                job['job_start_time'] >= latest_job['job_start_time']
            )
    if mark_as_latest:
        latest_path = paths.get_job_latest_path(
            dashboard=job['dashboard_name'],
            data_dir=data_dir,
        )
        latest_path_tmp = latest_path + '_tmp'
        with open(latest_path_tmp, 'w') as f:
            json.dump(job, f)
        shutil.move(latest_path_tmp, latest_path)



"""perform dashboard management actions"""

from __future__ import annotations

import typing

import mince.dashboards

if typing.TYPE_CHECKING:
    from typing import Type
    from mince import types


def run_dashboard(
    dashboard: str | Type[mince.dashboards.Dashboard],
    *,
    instance_kwargs: typing.Mapping[str, typing.Any] | None = None,
    run_kwargs: typing.Mapping[str, typing.Any] | None = None,
    registry: types.RegistryReference = None,
) -> None:
    instance = mince.dashboards.instantiate_dashboard(
        dashboard=dashboard,
        registry=registry,
        instance_kwargs=instance_kwargs,
    )

    # run dashboard
    if run_kwargs is None:
        run_kwargs = {}
    instance.run(**run_kwargs)


def find_available_port(
    min_port: int | str = 8052, *, n_attempts: int = 100
) -> int:
    import socket

    port = int(min_port)
    while True:
        try:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.bind(('localhost', port))
                return port
        except OSError:
            port += 1
